from collections import Counter, defaultdict
from copy import deepcopy
from math import ceil
from typing import List, Optional, Tuple

import numpy as np

from lhotse import CutSet
from lhotse.cut import Cut
from lhotse.utils import Seconds, TimeSpan, ifnone, is_module_available


class CutSetStatistics:
    """
    Low-level utility for creating an overview human-readable description for a CutSet.
    It powers the :meth:`~lhotse.cut.set.CutSet.describe` utility. Unlike `describe`,
    it allows to gather stats for multiple cut sets in parallel and combine and display them later.

    Example workflow (typically, the loop in the second line would be parallelized)::

        >>> cut_sets = [CutSet(...), CutSet(...)]
        >>> stats = [CutSetStatistics().accumulate(cuts) for cuts in cut_sets]
        >>> total_stats = stats[0].combine(*stats[1:])
        >>> total_stats.describe()
    """

    def __init__(self, full: bool = False):
        if not is_module_available("tabulate"):
            raise ValueError(
                "Since Lhotse v1.11, this function requires the `tabulate` package to be "
                "installed. Please run 'pip install tabulate' to continue."
            )

        self.full = full
        self.counters = defaultdict(int)
        self.cut_custom, self.sup_custom = Counter(), Counter()
        self.cut_durations = []
        self.speaking_time_durations, self.speech_durations = [], []

        if self.full:
            self.durations_by_num_speakers = defaultdict(list)
            self.single_durations, self.overlapped_durations = [], []

    def combine(self, *other: "CutSetStatistics") -> "CutSetStatistics":
        """Combinemultiple statistics into a new statistics object (does not modify self)."""
        lhs = deepcopy(self)
        for rhs in other:

            assert (
                lhs.full == rhs.full
            ), "Cannot combine statistics gathered with full=True and full=False."

            # Update Dict[str, Number] types
            for attr in ("counters", "cut_custom", "sup_custom"):
                for k in getattr(lhs, attr):
                    getattr(lhs, attr)[k] += getattr(rhs, attr)[k]

            # Update List[Any] types
            for attr in (
                "cut_durations",
                "speaking_time_durations",
                "speech_durations",
            ) + (("single_durations", "overlapped_durations") if lhs.full else ()):
                getattr(lhs, attr).extend(getattr(rhs, attr))

            if lhs.full:
                for k in lhs.durations_by_num_speakers:
                    lhs.durations_by_num_speakers[k].extend(
                        rhs.durations_by_num_speakers[k]
                    )

        return lhs

    def accumulate(self, cuts: CutSet) -> "CutSetStatistics":
        """Gather statistics to display later for a cut set."""

        def total_duration_(segments: List[TimeSpan]) -> float:
            return sum(segment.duration for segment in segments)

        for c in cuts:
            self.cut_durations.append(c.duration)
            if hasattr(c, "custom"):
                for key in ifnone(c.custom, ()):
                    self.cut_custom[key] += 1
            self.counters["recordings"] += int(c.has_recording)
            self.counters["features"] += int(c.has_features)

            # Total speaking time duration is computed by summing the duration of all
            # supervisions in the cut.
            for s in c.trimmed_supervisions:
                self.speaking_time_durations.append(s.duration)
                self.counters["supervisions"] += 1
                for key in ifnone(s.custom, ()):
                    self.sup_custom[key] += 1

            # Total speech duration is the sum of intervals where 1 or more speakers are
            # active.
            self.speech_durations.append(
                total_duration_(find_segments_with_speaker_count(c, min_speakers=1))
            )

            if self.full:
                # Duration of single-speaker segments
                self.single_durations.append(
                    total_duration_(
                        find_segments_with_speaker_count(
                            c, min_speakers=1, max_speakers=1
                        )
                    )
                )
                # Duration of overlapped segments
                self.overlapped_durations.append(
                    total_duration_(
                        find_segments_with_speaker_count(
                            c, min_speakers=2, max_speakers=None
                        )
                    )
                )
                # Durations by number of speakers (we assume that overlaps can happen between
                # at most 4 speakers. This is a reasonable assumption for most datasets.)
                self.durations_by_num_speakers[1].append(self.single_durations[-1])
                for num_spk in range(2, 5):
                    self.durations_by_num_speakers[num_spk].append(
                        total_duration_(
                            find_segments_with_speaker_count(
                                c, min_speakers=num_spk, max_speakers=num_spk
                            )
                        )
                    )

        return self

    def describe(self) -> None:
        """Display accumulated statistics in the CLI."""
        from tabulate import tabulate

        def convert_(seconds: Seconds) -> Tuple[int, int, int]:
            hours, seconds = divmod(seconds, 3600)
            minutes, seconds = divmod(seconds, 60)
            return int(hours), int(minutes), ceil(seconds)

        def time_as_str_(seconds: Seconds) -> str:
            h, m, s = convert_(seconds)
            return f"{h:02d}:{m:02d}:{s:02d}"

        cut_durations = self.cut_durations

        total_sum = np.array(cut_durations).sum()

        cut_stats = []
        cut_stats.append(["Cuts count:", len(cut_durations)])
        cut_stats.append(["Total duration (hh:mm:ss)", time_as_str_(total_sum)])
        cut_stats.append(["mean", f"{np.mean(cut_durations):.1f}"])
        cut_stats.append(["std", f"{np.std(cut_durations):.1f}"])
        cut_stats.append(["min", f"{np.min(cut_durations):.1f}"])
        cut_stats.append(["25%", f"{np.percentile(cut_durations, 25):.1f}"])
        cut_stats.append(["50%", f"{np.median(cut_durations):.1f}"])
        cut_stats.append(["75%", f"{np.percentile(cut_durations, 75):.1f}"])
        cut_stats.append(["99%", f"{np.percentile(cut_durations, 99):.1f}"])
        cut_stats.append(["99.5%", f"{np.percentile(cut_durations, 99.5):.1f}"])
        cut_stats.append(["99.9%", f"{np.percentile(cut_durations, 99.9):.1f}"])
        cut_stats.append(["max", f"{np.max(cut_durations):.1f}"])

        for key, val in self.counters.items():
            cut_stats.append([f"{key.title()} available:", val])

        print("Cut statistics:")
        print(tabulate(cut_stats, tablefmt="fancy_grid"))

        if self.cut_custom:
            print("CUT custom fields:")
            for key, val in self.cut_custom.most_common():
                print(f"- {key} (in {val} cuts)")

        if self.sup_custom:
            print("SUPERVISION custom fields:")
            for key, val in self.sup_custom.most_common():
                print(f"- {key} (in {val} cuts)")

        total_speech = np.array(self.speech_durations).sum()
        total_speaking_time = np.array(self.speaking_time_durations).sum()
        total_silence = total_sum - total_speech
        speech_stats = []
        speech_stats.append(
            [
                "Total speech duration",
                time_as_str_(total_speech),
                f"{total_speech / total_sum:.2%} of recording",
            ]
        )
        speech_stats.append(
            [
                "Total speaking time duration",
                time_as_str_(total_speaking_time),
                f"{total_speaking_time / total_sum:.2%} of recording",
            ]
        )
        speech_stats.append(
            [
                "Total silence duration",
                time_as_str_(total_silence),
                f"{total_silence / total_sum:.2%} of recording",
            ]
        )
        if self.full:
            total_single = np.array(self.single_durations).sum()
            total_overlap = np.array(self.overlapped_durations).sum()
            speech_stats.append(
                [
                    "Single-speaker duration",
                    time_as_str_(total_single),
                    f"{total_single / total_sum:.2%} ({total_single / total_speech:.2%} of speech)",
                ]
            )
            speech_stats.append(
                [
                    "Overlapped speech duration",
                    time_as_str_(total_overlap),
                    f"{total_overlap / total_sum:.2%} ({total_overlap / total_speech:.2%} of speech)",
                ]
            )
        print("Speech duration statistics:")
        print(tabulate(speech_stats, tablefmt="fancy_grid"))

        if not self.full:
            return

        # Additional statistics for full report
        speaker_stats = [
            [
                "Number of speakers",
                "Duration (hh:mm:ss)",
                "Speaking time (hh:mm:ss)",
                "% of speech",
                "% of speaking time",
            ]
        ]
        for num_spk, durations in self.durations_by_num_speakers.items():
            speaker_sum = np.array(durations).sum()
            speaking_time = num_spk * speaker_sum
            speaker_stats.append(
                [
                    num_spk,
                    time_as_str_(speaker_sum),
                    time_as_str_(speaking_time),
                    f"{speaker_sum / total_speech:.2%}",
                    f"{speaking_time / total_speaking_time:.2%}",
                ]
            )

        speaker_stats.append(
            [
                "Total",
                time_as_str_(total_speech),
                time_as_str_(total_speaking_time),
                "100.00%",
                "100.00%",
            ]
        )

        print("Speech duration statistics by number of speakers:")
        print(tabulate(speaker_stats, headers="firstrow", tablefmt="fancy_grid"))


def find_segments_with_speaker_count(
    cut: Cut, min_speakers: int = 0, max_speakers: Optional[int] = None
) -> List[TimeSpan]:
    """
    Given a Cut, find a list of intervals that contain the specified number of speakers.

    :param cuts: the Cut to search.
    :param min_speakers: the minimum number of speakers.
    :param max_speakers: the maximum number of speakers.
    :return: a list of TimeSpans.
    """
    if max_speakers is None:
        max_speakers = float("inf")

    assert (
        min_speakers >= 0 and min_speakers <= max_speakers
    ), f"min_speakers={min_speakers} and max_speakers={max_speakers} are not valid."

    # First take care of trivial cases.
    if min_speakers == 0 and max_speakers == float("inf"):
        return [TimeSpan(0, cut.duration)]
    if len(cut.supervisions) == 0:
        return [] if min_speakers > 0 else [TimeSpan(0, cut.duration)]

    # We collect all the timestamps of the supervisions in the cut. Each timestamp is
    # a tuple of (time, is_speaker_start).
    timestamps = []
    # Add timestamp for cut start
    timestamps.append((0.0, None))
    for segment in cut.supervisions:
        timestamps.append((segment.start, True))
        timestamps.append((segment.end, False))
    # Add timestamp for cut end
    timestamps.append((cut.duration, None))

    # Sort the timestamps. We need the following priority order:
    # 1. Time mark of the timestamp: lower time mark comes first.
    # 2. For timestamps with the same time mark, None < False < True.
    timestamps.sort(key=lambda x: (x[0], x[1] is not None, x[1] is True))

    # We remove the timestamps that are not relevant for the search. The desired range
    # is given by the range of the cut start and end timestamps.
    cut_start_idx, cut_end_idx = [i for i, t in enumerate(timestamps) if t[1] is None]
    timestamps = timestamps[cut_start_idx : cut_end_idx + 1]

    # Now we iterate over the timestamps and count the number of speakers in any
    # given time interval. If the number of speakers is in the desired range,
    # we keep the interval.
    num_speakers = 0
    seg_start = 0.0
    intervals = []
    for timestamp, is_start in timestamps[1:]:
        if num_speakers >= min_speakers and num_speakers <= max_speakers:
            intervals.append((seg_start, timestamp))
        if is_start is not None:
            num_speakers += 1 if is_start else -1
        seg_start = timestamp

    # Merge consecutive intervals and remove empty intervals.
    merged_intervals = []
    for start, end in intervals:
        if start == end:
            continue
        if merged_intervals and merged_intervals[-1][1] == start:
            merged_intervals[-1] = (merged_intervals[-1][0], end)
        else:
            merged_intervals.append((start, end))

    return [TimeSpan(start, end) for start, end in merged_intervals]



import abc
import logging
import multiprocessing
from concurrent.futures import ProcessPoolExecutor
from functools import partial, partialmethod
from itertools import chain
from typing import Dict, Generator, List, Optional, Tuple, Type, Union

import torch
from tqdm.auto import tqdm

from lhotse.cut import Cut, CutSet
from lhotse.supervision import AlignmentItem
from lhotse.utils import fastcopy


class FailedToAlign(RuntimeError):
    pass


class ForcedAligner(abc.ABC):
    """
    ForcedAligner is an abstract base class for forced aligners.
    """

    def __init__(self, device: Union[str, torch.device] = "cpu"):
        self.device = torch.device(device)

    @abc.abstractproperty
    def sample_rate(self) -> int:
        pass

    @abc.abstractmethod
    def normalize_text(
        self, text: str, language: Optional[str] = None
    ) -> Union[str, List[str]]:
        pass

    @abc.abstractmethod
    def align(
        self, audio: torch.Tensor, transcript: Union[str, List[Tuple[str, str]]]
    ) -> List[AlignmentItem]:
        pass

    def __call__(self, cut: Cut, normalize: bool = True) -> Cut:
        cut = fastcopy(cut, supervisions=list(cut.supervisions))

        for idx, subcut in enumerate(cut.trim_to_supervisions(keep_overlapping=False)):
            audio = torch.as_tensor(
                subcut.resample(self.sample_rate).load_audio(), device=self.device
            )
            sup = subcut.supervisions[0]
            transcript = (
                self.normalize_text(sup.text, language=sup.language)
                if normalize
                else sup.text
            )

            try:
                pre_alignment = self.align(audio, transcript)
            except FailedToAlign:
                logging.exception(
                    f"Failed to align supervision '{sup.id}' for cut '{cut.id}'. Writing it without alignment."
                )
                continue

            alignment = [
                item._replace(start=item.start + subcut.start) for item in pre_alignment
            ]

            # Important: reference the original supervision before "trim_to_supervisions"
            #            because the new one has start=0 to match the start of the subcut
            sup = cut.supervisions[idx].with_alignment(kind="word", alignment=alignment)
            cut.supervisions[idx] = sup

        return cut



from pathlib import Path
from typing import Optional, TypeVar, Union

from lhotse import AudioSource, Features, Recording, compute_num_samples, fastcopy
from lhotse.array import Array, TemporalArray
from lhotse.cut import Cut
from lhotse.utils import Pathlike

Manifest = TypeVar("Manifest", Recording, Features, Array, TemporalArray)


def to_shar_placeholder(manifest: Manifest, cut: Optional[Cut] = None) -> Manifest:
    if isinstance(manifest, Recording):
        return fastcopy(
            manifest,
            # Creates a single AudioSource out of multiple ones.
            sources=[
                AudioSource(type="shar", channels=manifest.channel_ids, source="")
            ],
            # Removes the transform metadata because they were already executed.
            transforms=None,
            duration=cut.duration if cut is not None else manifest.duration,
            num_samples=compute_num_samples(cut.duration, manifest.sampling_rate)
            if cut is not None
            else manifest.num_samples,
        )
    elif isinstance(manifest, Array):
        return fastcopy(manifest, storage_type="shar", storage_path="", storage_key="")
    elif isinstance(manifest, Features):
        return fastcopy(
            manifest,
            start=0,
            duration=cut.duration if cut is not None else manifest.duration,
            storage_type="shar",
            storage_path="",
            storage_key="",
        )
    elif isinstance(manifest, TemporalArray):
        return fastcopy(
            manifest,
            start=0,
            array=fastcopy(
                manifest.array,
                storage_type="shar",
                storage_path="",
                storage_key="",
            ),
        )
    else:
        raise RuntimeError(f"Unexpected manifest type: {type(manifest)}")


def fill_shar_placeholder(
    manifest: Union[Cut, Recording, Features, Array, TemporalArray],
    data: bytes,
    tarpath: Pathlike,
    field: Optional[str] = None,
) -> None:
    if isinstance(manifest, Cut):
        assert (
            field is not None
        ), "'field' argument must be provided when filling a Shar placeholder in a Cut."
        manifest = getattr(manifest, field)
        fill_shar_placeholder(
            manifest=manifest, field=field, data=data, tarpath=tarpath
        )

    tarpath = Path(tarpath)

    if isinstance(manifest, Recording):
        assert (
            len(manifest.sources) == 1
        ), "We expected a single (possibly multi-channel) AudioSource in Shar format."
        manifest.sources[0].type = "memory"
        manifest.sources[0].source = data

    elif isinstance(manifest, (Features, Array)):
        manifest.storage_key = data
        if tarpath.suffix == ".llc":
            manifest.storage_type = "memory_lilcom"
        elif tarpath.suffix == ".npy":
            manifest.storage_type = "memory_npy"
        else:
            raise RuntimeError(f"Unknown array/tensor format: {tarpath}")

    elif isinstance(manifest, TemporalArray):
        manifest.array.storage_key = data
        if tarpath.suffix == ".llc":
            manifest.array.storage_type = "memory_lilcom"
        elif tarpath.suffix == ".npy":
            manifest.array.storage_type = "memory_npy"
        else:
            raise RuntimeError(f"Unknown array/tensor format: {tarpath}")

    else:
        raise RuntimeError(f"Unknown manifest type: {type(manifest).__name__}")



from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Optional, Tuple

import numpy as np

from lhotse.utils import Pathlike, fastcopy, is_module_available


@dataclass
class Image:
    """
    The Image manifest describes an image that is stored somewhere: it might be
    in a file, in memory, in the cloud, etc.
    Image helps abstract away from the actual storage mechanism and location by
    providing a method called :meth:`.Image.load`.

    Image manifest can be easily created by calling
    :meth:`lhotse.image.io.PillowWriter.store_image`, for example::

        >>> from lhotse.image.io import PillowWriter
        >>> with PillowWriter('images/') as writer:
        ...     manifest = writer.store_image('image-1', 'path/to/image.jpg')
    """

    # Storage type defines which image reader type should be instantiated
    # e.g. 'pillow_files', 'pillow_memory'
    storage_type: str

    # Storage path is either the path to a directory holding files with images
    # or None for in-memory images
    storage_path: str

    # Storage key is the name of the file in a directory
    # or binary data for in-memory images
    storage_key: str

    # Width and height of the image in pixels
    width: int
    height: int

    @property
    def shape(self) -> Tuple[int, int]:
        """Return the shape of the image as (height, width)."""
        return (self.height, self.width)

    @property
    def is_in_memory(self) -> bool:
        """Check if the image is stored in memory."""
        # Import locally to avoid circular dependencies
        from lhotse.image.io import is_in_memory

        return is_in_memory(self.storage_type)

    def to_dict(self) -> dict:
        """Convert to dictionary for serialization."""
        return asdict(self)

    @classmethod
    def from_dict(cls, data: dict) -> "Image":
        """Create an Image from a dictionary representation."""
        if (
            "storage_key" in data
            and "storage_type" in data
            and "storage_path" not in data
        ):
            data["storage_path"] = None
        return cls(**data)

    @classmethod
    def from_pillow(cls, data) -> "Image":
        from lhotse.image.io import PillowInMemoryWriter

        return PillowInMemoryWriter().store_image("", data)

    def load(self, as_pil_image: bool = False):
        """
        Load the image from the underlying storage.

        :param as_numpy: If True (default), returns a numpy array with shape (height, width, channels).
                         If False, returns a PIL Image object.
        :return: Loaded image as numpy array or PIL Image object.
        """
        assert is_module_available(
            "PIL"
        ), "In order to load images, please run 'pip install pillow'"
        # Import locally to avoid circular dependencies
        from lhotse.image.io import get_reader

        # noinspection PyArgumentList
        storage = get_reader(self.storage_type)(self.storage_path)
        # Load the image from the storage
        img_data = storage.read(self.storage_key, as_pil_image=as_pil_image)
        return img_data

    def with_path_prefix(self, path: Pathlike) -> "Image":
        """
        Return a copy of the image with ``path`` added as a prefix
        to the ``storage_path`` member.
        """
        return fastcopy(self, storage_path=str(Path(path) / self.storage_path))

    def move_to_memory(self) -> "Image":
        """
        Return a copy of the image stored in memory.
        """
        # Import locally to avoid circular dependencies
        from lhotse.image.io import get_memory_writer

        if self.storage_type in ("pillow_memory"):
            return self  # nothing to do

        img = self.load()
        writer = get_memory_writer("pillow_memory")()
        data = writer.write("", img)  # key is ignored by in memory writers
        return Image(
            storage_type=writer.name,
            storage_key=data,
            storage_path="",
            width=self.width,
            height=self.height,
        )

    def __repr__(self):
        return (
            f"Image("
            f"storage_type='{self.storage_type}', "
            f"storage_path='{self.storage_path}', "
            f"storage_key='{self.storage_key if isinstance(self.storage_key, str) else '<binary-data>'}', "
            f"width={self.width}, "
            f"height={self.height}"
            f")"
        )

    def plot(self, ax=None, **kwargs):
        """
        Display the image using matplotlib.

        :param ax: Optional matplotlib Axes to plot on. If None, a new figure and axes will be created.
        :param kwargs: Additional keyword arguments passed to `ax.imshow()`.
        :return: The matplotlib Axes object with the plotted image.
        """
        if not is_module_available("matplotlib"):
            raise ImportError(
                "To use Image.plot(), please 'pip install matplotlib' first."
            )
        import matplotlib.pyplot as plt

        img_data = self.load()

        if ax is None:
            fig, ax = plt.subplots()

        ax.imshow(img_data, **kwargs)
        ax.axis("off")  # Typically, we don't want axis ticks for images
        return ax



"""
The CHiME-6 dataset is a collection of over 50 hours of conversational speech recordings
collected from twenty real dinner parties that have taken place in real homes. The
recordings have been made using multiple 4-channel microphone arrays and have been
fully transcribed.

The dataset features:

- simultaneous recordings from multiple microphone arrays;
- real conversation, i.e. talkers speaking in a relaxed and unscripted fashion;
- a range of room acoustics from 20 different homes each with two or three separate
  recording areas;
- real domestic noise backgrounds, e.g., kitchen appliances, air conditioning,
  movement, etc.

Fully-transcribed utterances are provided in continuous audio with ground truth speaker
labels and start/end time annotations for segmentation.

The dataset was used for the 5th and 6th CHiME Speech Separation and Recognition
Challenge. Further information and an open source baseline speech recognition system
are available online (http://spandh.dcs.shef.ac.uk/chime_challenge/chime2018).

NOTE: The CHiME-5 and CHiME-6 datasets are the same, with the only difference that
additional software was provided in CHiME-6 to perform array synchronization. We expect
that users have already downloaded the CHiME-5 dataset here:
https://licensing.sheffield.ac.uk/product/chime5

NOTE: Users can also additionally perform array synchronization as described here:
https://github.com/kaldi-asr/kaldi/blob/master/egs/chime6/s5_track1/local/generate_chime6_data.sh
We also provide this option in the `prepare_chime6` function.
"""

import itertools
import json
import logging
import os
import shutil
import subprocess
import tempfile
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor, wait
from datetime import datetime as dt
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Union

import numpy as np
from tqdm import tqdm

from lhotse import fix_manifests, validate_recordings_and_supervisions
from lhotse.audio import AudioSource, Recording, RecordingSet
from lhotse.recipes.utils import TimeFormatConverter, normalize_text_chime6
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, add_durations, resumable_download

# fmt: off
DATASET_PARTS = {
    "train": [
        "S03", "S04", "S05", "S06", "S07", "S08", "S12", "S13",
        "S16", "S17", "S18", "S19", "S20", "S22", "S23", "S24",
    ],
    "dev": ["S02", "S09"],
    "eval": ["S01", "S21"],
}

DATASET_PARTS_CHIME7 = {
    "train": [
        "S03", "S04", "S05", "S06", "S07", "S08", "S12", "S13",
        "S16", "S17", "S18", "S22", "S23", "S24",
    ],
    "dev": ["S02", "S09"],
    "eval": ["S01", "S19", "S20", "S21"],
}

# fmt: on
CHIME6_AUDIO_EDITS_JSON = "https://raw.githubusercontent.com/chimechallenge/chime6-synchronisation/master/chime6_audio_edits.json"
CHIME6_MD5SUM_FILE = "https://raw.githubusercontent.com/chimechallenge/chime6-synchronisation/master/audio_md5sums.txt"


def download_chime6(
    target_dir: Pathlike = ".",
) -> Path:
    """
    Download the original dataset. This cannot be done automatically because of the
    license agreement. Please visit the following URL and download the dataset manually:
    https://licensing.sheffield.ac.uk/product/chime5
    :param target_dir: Pathlike, the path of the dir to storage the dataset.
    :return: the path to downloaded and extracted directory with data.
    """
    print("We cannot download the CHiME-6 dataset automatically.")
    print("Please visit the following URL and download the dataset manually:")
    print("https://licensing.sheffield.ac.uk/product/chime5")
    print("Then, please extract the tar files to the following directory:")
    print(f"{target_dir}")
    return Path(target_dir)


def prepare_chime6(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
    dataset_parts: Optional[Union[str, Sequence[str]]] = "all",
    mic: str = "mdm",
    use_reference_array: bool = False,
    perform_array_sync: bool = False,
    verify_md5_checksums: bool = False,
    num_jobs: int = 1,
    num_threads_per_job: int = 1,
    sox_path: Pathlike = "/usr/bin/sox",
    normalize_text: str = "kaldi",
    use_chime7_split: bool = False,
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions
    :param corpus_dir: Pathlike, the path of the data dir, either the original CHiME-5
        data or the synchronized CHiME-6 data. If former, the `perform_array_sync`
        must be True.
    :param output_dir: Pathlike, the path where to write the manifests.
    :param mic: str, the microphone type to use, choose from "ihm" (close-talk) or "mdm"
        (multi-microphone array) settings. For MDM, there are 6 array devices with 4
        channels each, so the resulting recordings will have 24 channels.
    :param use_reference_array: bool, if True, use the reference array for MDM setting.
        Only the supervision segments have the reference array information in the
        `channel` field. The recordings will still have all the channels in the array.
        Note that the train set does not have the reference array information.
    :param perform_array_sync: Bool, if True, perform array synchronization based on:
        https://github.com/chimechallenge/chime6-synchronisation
    :param num_jobs: int, the number of jobs to run in parallel for array synchronization.
    :param num_threads_per_job: int, number of threads to use per job for clock drift
        correction. Large values may require more memory, so we recommend using a job
        scheduler.
    :param sox_path: Pathlike, the path to the sox v14.4.2 binary. Note that different
        versions of sox may produce different results.
    :param normalize_text: str, the text normalization method, choose from "none", "upper",
        "kaldi". The "kaldi" method is the same as Kaldi's text normalization method for
        CHiME-6.
    :param verify_md5_checksums: bool, if True, verify the md5 checksums of the audio files.
        Note that this step is slow so we recommend only doing it once. It can be sped up
        by using the `num_jobs` argument.
    :param use_chime7_split: bool, if True, use the new split for CHiME-7 challenge.
    :return: a Dict whose key is the dataset part ("train", "dev" and "eval"), and the
        value is Dicts with the keys 'recordings' and 'supervisions'.

    NOTE: If `perform_array_sync` is True, the synchronized data will be written to
        `output_dir`/CHiME6. This may take a long time and the output will occupy
        approximately 160G of storage. We will also create a temporary directory for
        processing, so the required storage in total will be approximately 300G.
    """
    import soundfile as sf

    assert mic in ["ihm", "mdm"], "mic must be either 'ihm' or 'mdm'."

    if output_dir:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

    if "all" in dataset_parts:
        dataset_parts = list(DATASET_PARTS.keys())
    elif isinstance(dataset_parts, str):
        dataset_parts = [dataset_parts]
    assert set(dataset_parts).issubset(
        set(DATASET_PARTS.keys())
    ), f"dataset_parts must be one of {list(DATASET_PARTS.keys())}. Found {dataset_parts}"

    sessions = list(
        itertools.chain.from_iterable([DATASET_PARTS[part] for part in dataset_parts])
    )

    if perform_array_sync:
        if not output_dir:
            raise ValueError(
                "If `perform_array_sync` is True, `output_dir` must be specified."
            )
        chime6_dir = output_dir / "CHiME6"
        chime6_dir.mkdir(parents=True, exist_ok=True)

        # Create directories for train, dev, and test audio and transcriptions
        for part in dataset_parts:
            (chime6_dir / "audio" / part).mkdir(parents=True, exist_ok=True)
            (chime6_dir / "transcriptions" / part).mkdir(parents=True, exist_ok=True)

        # Check sox version
        sox_version = (
            subprocess.check_output([sox_path, "--version"])
            .decode("utf-8")
            .strip()
            .split(" ")[-1]
        )
        assert sox_version == "v14.4.2", (
            "The sox version must be 14.4.2. "
            "Please download the sox v14.4.2 binary from "
            "https://sourceforge.net/projects/sox/files/sox/14.4.2/ "
            "and specify the path to the binary with the `sox_path` argument."
            "You can also install it in a Conda environment with the following command: "
            "conda install -c conda-forge sox=14.4.2"
        )

        chime6_array_synchronizer = Chime6ArraySynchronizer(
            corpus_dir=corpus_dir,
            output_dir=chime6_dir,
            sox_path=sox_path,
            num_workers=num_threads_per_job,
        )

        num_jobs = min(num_jobs, len(sessions))  # since there are 20 sessions
        logging.info(
            f"Performing array synchronization with {num_jobs} jobs. This may "
            "take a long time."
        )
        with ProcessPoolExecutor(max_workers=num_jobs) as ex:
            futures = [
                ex.submit(
                    chime6_array_synchronizer.synchronize_session,
                    session=session,
                )
                for session in sessions
            ]
            _ = wait(futures)
    else:
        chime6_dir = Path(corpus_dir)

    # Verify MD5 checksums for all audio files if requested
    if verify_md5_checksums:
        if _verify_md5_checksums(chime6_dir, num_workers=num_jobs, sessions=sessions):
            print("MD5 checksums verified. All OK.")
        else:
            raise RuntimeError(
                "MD5 checksums do not match. Please prepare the array-synchronized CHiME-6 "
                "dataset again."
            )

    # Reference array is only applicable for MDM setting
    use_reference_array = use_reference_array and mic == "mdm"

    manifests = defaultdict(dict)

    for part in dataset_parts:
        recordings = []
        supervisions = []

        # Since CHiME-7 uses a different split, we need to change the sessions
        if use_chime7_split:
            DATASET_PARTS[part] = DATASET_PARTS_CHIME7[part]
        # Also, if the session is S19 or S20, we will look for its audio and transcriptions
        # in the train set, since it was originally in train.

        # First we create the recordings
        if mic == "ihm":
            global_spk_channel_map = {}
            for session in DATASET_PARTS[part]:
                part_ = (
                    "train" if use_chime7_split and session in ["S19", "S20"] else part
                )

                audio_paths = [
                    p for p in (chime6_dir / "audio" / part_).rglob(f"{session}_P*.wav")
                ]

                if len(audio_paths) == 0:
                    raise FileNotFoundError(
                        f"No audio found for session {session} in {part_} set."
                    )

                sources = []
                # NOTE: Each headset microphone is binaural
                for idx, audio_path in enumerate(audio_paths):
                    sources.append(
                        AudioSource(
                            type="file",
                            channels=[2 * idx, 2 * idx + 1],
                            source=str(audio_path),
                        )
                    )
                    spk_id = audio_path.stem.split("_")[1]
                    global_spk_channel_map[(session, spk_id)] = [2 * idx, 2 * idx + 1]

                audio_sf = sf.SoundFile(str(audio_paths[0]))

                recordings.append(
                    Recording(
                        id=session,
                        sources=sources,
                        sampling_rate=int(audio_sf.samplerate),
                        num_samples=audio_sf.frames,
                        duration=audio_sf.frames / audio_sf.samplerate,
                    )
                )

        else:
            for session in DATASET_PARTS[part]:
                part_ = (
                    "train" if use_chime7_split and session in ["S19", "S20"] else part
                )

                audio_paths = [
                    p for p in (chime6_dir / "audio" / part_).rglob(f"{session}_U*.wav")
                ]

                sources = []
                for idx, audio_path in enumerate(sorted(audio_paths)):
                    sources.append(
                        AudioSource(type="file", channels=[idx], source=str(audio_path))
                    )

                audio_sf = sf.SoundFile(str(audio_paths[0]))

                recordings.append(
                    Recording(
                        id=session,
                        sources=sources,
                        sampling_rate=int(audio_sf.samplerate),
                        num_samples=audio_sf.frames,
                        duration=audio_sf.frames / audio_sf.samplerate,
                    )
                )

        recordings = RecordingSet.from_recordings(recordings)

        def _get_channel(spk_id, session, ref=None):
            if mic == "ihm":
                return global_spk_channel_map[(session, spk_id)]
            else:
                recording = recordings[session]
                return (
                    list(range(recording.num_channels))
                    if not ref
                    else [i for i, s in enumerate(recording.sources) if ref in s.source]
                )

        # Then we create the supervisions
        for session in DATASET_PARTS[part]:
            part_ = "train" if use_chime7_split and session in ["S19", "S20"] else part

            with open(chime6_dir / "transcriptions" / part_ / f"{session}.json") as f:
                transcript = json.load(f)
                for idx, segment in enumerate(transcript):
                    spk_id = segment["speaker"]
                    channel = _get_channel(
                        spk_id,
                        session,
                        ref=segment["ref"]
                        if use_reference_array and part != "train"
                        else None,
                    )
                    start = TimeFormatConverter.hms_to_seconds(segment["start_time"])
                    end = TimeFormatConverter.hms_to_seconds(segment["end_time"])
                    if start >= end:  # some segments may have negative duration
                        continue
                    supervisions.append(
                        SupervisionSegment(
                            id=f"{session}-{idx}",
                            recording_id=session,
                            start=start,
                            duration=add_durations(end, -start, sampling_rate=16000),
                            channel=channel,
                            text=normalize_text_chime6(
                                segment["words"], normalize=normalize_text
                            ),
                            language="English",
                            speaker=spk_id,
                            custom={
                                "location": segment["location"],
                            }
                            if part != "train" and "location" in segment
                            else None,
                        )
                    )

        supervisions = SupervisionSet.from_segments(supervisions)

        recording_set, supervision_set = fix_manifests(
            recordings=recordings, supervisions=supervisions
        )
        # Fix manifests
        validate_recordings_and_supervisions(recording_set, supervision_set)

        if output_dir is not None:
            mic_affix = f"{mic}-ref" if use_reference_array else mic
            supervision_set.to_file(
                output_dir / f"chime6-{mic_affix}_supervisions_{part}.jsonl.gz"
            )
            recording_set.to_file(
                output_dir / f"chime6-{mic}_recordings_{part}.jsonl.gz"
            )

        manifests[part] = {
            "recordings": recording_set,
            "supervisions": supervision_set,
        }

    return manifests


def _verify_md5_checksums(
    corpus_dir: Pathlike, num_workers: int = 1, sessions: Optional[List[str]] = None
) -> bool:
    import hashlib

    # First download checksum file and read it into a dictionary
    temp_dir = Path(tempfile.mkdtemp())
    checksum_file = temp_dir / "md5sums.txt"
    resumable_download(
        CHIME6_MD5SUM_FILE, str(checksum_file), desc="Downloading checksum file"
    )
    checksums = {}
    with open(checksum_file, "r") as f:
        for line in f:
            checksum, filename = line.strip().split(" ", maxsplit=1)
            checksums[Path(filename).stem] = checksum

    # Now verify the checksums
    def _verify_checksum(file: Pathlike) -> bool:
        checksum = hashlib.md5(open(str(file), "rb").read()).hexdigest()
        filename = str(file.stem)
        if filename in checksums and checksum != checksums[filename]:
            return False
        return True

    all_files = list(corpus_dir.rglob("*.wav"))
    if sessions is not None:
        all_files = [f for f in all_files if f.stem.split("_")[0] in sessions]

    print(f"Verifying checksum with {num_workers} workers...")
    with ThreadPoolExecutor(max_workers=num_workers) as ex:
        results = list(tqdm(ex.map(_verify_checksum, all_files), total=len(all_files)))

    return all(results)


# The following class is based on functions from:
# https://github.com/chimechallenge/chime6-synchronisation
# We have made 2 changes to get some speed-up:
# 1. We combine all channels in an array for applying frame drop correction.
# 2. We apply multi-threading with 4 threads for clock drift correction.


class Chime6ArraySynchronizer:
    """
    Class for synchronizing CHiME6 array recordings.
    """

    def __init__(
        self,
        corpus_dir: Pathlike,
        output_dir: Pathlike,
        sox_path: Pathlike = "sox",
        num_workers: int = 1,
    ) -> None:
        self.corpus_dir = Path(corpus_dir)
        self.output_dir = Path(output_dir)
        self.sox_path = Path(sox_path)
        self.num_workers = num_workers

        # Create output directory
        self.output_dir.mkdir(exist_ok=True, parents=True)

        # Download the audio edits JSON file
        audio_edits_json = self.output_dir / "audio_edits.json"
        resumable_download(CHIME6_AUDIO_EDITS_JSON, str(audio_edits_json))
        with open(audio_edits_json) as f:
            self.audio_edits = dict(json.load(f))

        # Check SOX path
        if not str(self.sox_path).endswith("sox"):
            self.sox_path = self.sox_path / "sox"

    def synchronize_session(self, session: str) -> None:
        """
        Synchronize a single CHiME6 session.
        """
        temp_dir = Path(
            tempfile.mkdtemp(prefix=f"chime6_{session}_", dir=self.output_dir)
        )

        if session not in self.audio_edits:
            logging.warning(f"No audio edits found for session {session}")
            return

        session_audio_edits = self.audio_edits[session]

        print(f"Correcting {session} for frame drops...")
        self.correct_frame_drops(temp_dir, session, frame_drops=session_audio_edits)

        print(f"Correcting {session} for clock drift...")
        self.correct_clock_drift(
            temp_dir,
            session,
            linear_fit=session_audio_edits,
            num_threads=self.num_workers,
        )

        print(f"Adjusting timestamps in {session} JSON files...")
        self.adjust_json_timestamps(session, linear_fit=session_audio_edits)

        # clean up
        shutil.rmtree(temp_dir)

        return

    def correct_frame_drops(
        self,
        output_dir: Pathlike,
        session: str,
        frame_drops: Optional[Dict[str, Any]] = None,
    ) -> None:

        # For binaural recordings, we just create symbolic links to the original files
        session_binaural_wavs = sorted(
            (self.corpus_dir / "audio").rglob(f"{session}_P*.wav")
        )
        for wav in session_binaural_wavs:
            wav_relative_path = wav.relative_to(self.corpus_dir)
            wav_output_path = output_dir / wav_relative_path
            wav_output_path.parent.mkdir(exist_ok=True, parents=True)
            os.symlink(wav, wav_output_path)

        # For array recordings, we need to apply the edits. We first group the channels by
        # their corresponding array. For example, the wav names are like: S02_U01.CH2.wav
        session_array_wavs = sorted(
            (self.corpus_dir / "audio").rglob(f"{session}_U*.wav")
        )
        array_wavs = defaultdict(list)
        for wav in session_array_wavs:
            array_id = wav.stem.split(".")[0].split("_")[-1]
            array_wavs[array_id].append(wav)

        # Then we apply the edits to each array
        for array_id, wavs in array_wavs.items():
            if array_id not in frame_drops:
                logging.warning(
                    f"Array {array_id} in session {session} has no frame drops information."
                )
                continue

            in_wavs, out_wavs = [], []
            for wav in wavs:
                wav_relative_path = wav.relative_to(self.corpus_dir)
                wav_output_path = output_dir / wav_relative_path
                in_wavs.append(wav)
                out_wavs.append(wav_output_path)

            # Apply the edits to the wavs
            self._apply_edits_to_wav(in_wavs, out_wavs, frame_drops[array_id]["edits"])

        return

    def _apply_edits_to_wav(
        self, in_wavs: Pathlike, out_wavs: Pathlike, edits: List[List[int]]
    ) -> None:
        import soundfile as sf

        x = np.concatenate(
            [Recording.from_file(wav).load_audio() for wav in in_wavs], axis=0
        )
        # Pre-allocate space for editted signal
        max_space = edits[-1][2] + edits[-1][1] - edits[-1][0]
        x_new = np.zeros(shape=(x.shape[0], max_space), dtype=x.dtype)
        length_x = x.shape[1]

        for edit in edits:
            in_from = edit[0]
            in_to = min(edit[1], length_x)
            out_from = edit[2]
            out_to = out_from + in_to - in_from
            if in_from > length_x:
                break
            x_new[:, out_from - 1 : out_to] = x[:, in_from - 1 : in_to]

        # Trim off excess
        x_new = x_new[:, 0:out_to]

        # Write to file
        for i, wav in enumerate(out_wavs):
            sf.write(
                file=str(wav),
                data=np.expand_dims(x_new[i], axis=1),
                samplerate=16000,
                format="WAV",
            )
        return

    def correct_clock_drift(
        self,
        corpus_dir: Pathlike,
        session: str,
        linear_fit: Optional[Dict[str, Any]] = None,
        num_threads: int = 1,
    ) -> None:
        session_wavs = sorted((corpus_dir / "audio").rglob(f"{session}_*.wav"))
        with ThreadPoolExecutor(max_workers=num_threads) as ex:
            futures = []
            for wav in session_wavs:
                wav_relative_path = wav.relative_to(corpus_dir)
                wav_output_path = self.output_dir / wav_relative_path

                # Get binaural mic id or array id
                mic_id = wav.name.split(".")[0].split("_")[-1]

                if mic_id not in linear_fit:
                    logging.warning(
                        f"Channel {mic_id} in session {session} has no clock drift information."
                    )
                    continue

                futures.append(
                    ex.submit(
                        self._apply_clock_drift_correction,
                        wav,
                        wav_output_path,
                        linear_fit[mic_id],
                    )
                )

            for future in tqdm(futures):
                future.result()
        return

    def _apply_clock_drift_correction(
        self,
        in_wav: Pathlike,
        out_wav: Pathlike,
        linear_fit: Dict[str, Union[int, float, List]],
    ) -> None:
        speeds = linear_fit["speed"]
        padding = linear_fit["padding"]
        filename = in_wav.stem

        sox_cmd = str(self.sox_path)
        in_wav = str(in_wav)
        out_wav = str(out_wav)

        if isinstance(speeds, list):
            # multisegment fit - only needed for S01_U02 and S01_U05
            starts = padding
            ends = padding[1:] + [-1]  # -1 means end of signal
            command_concat = [sox_cmd]
            samples_to_lose = 0

            tmpdir = tempfile.mkdtemp(dir=self.output_dir)

            for seg, (start, end, speed) in enumerate(zip(starts, ends, speeds)):
                # print(seg, start, speed, end)
                of1 = tmpdir + "/" + filename + "." + str(seg) + ".wav"
                of2 = tmpdir + "/" + filename + ".x" + str(seg) + ".wav"

                command1 = [sox_cmd, "-D", "-R", in_wav, of1]
                if seg == 0:
                    # 'start' has a different meaning for first segment
                    # Acts like padding does in the simple one-piece case
                    if start < 0:
                        start = -start
                        trim = ["trim", f"{start}s"]
                    else:
                        trim = ["pad", f"{start}s", "0s", "trim", "0s"]
                else:
                    start += samples_to_lose  # may need to truncate some samples
                    trim = ["trim", f"{int(start)}s"]
                command1 += trim
                if end > 0:  # segment of given duration
                    duration = end - start
                    command1 += [f"{duration}s"]

                if speed < 0:
                    # Negative speed means backward in time,
                    # Effectively have to remove some samples
                    # This happen in S01_U05.
                    samples_to_lose = -duration / speed
                else:
                    samples_to_lose = 0
                    command2 = [sox_cmd, "-D", "-R", of1, of2, "speed", str(speed)]
                    subprocess.call(command1)
                    subprocess.call(command2)
                    command_concat.append(of2)

            command_concat.append(out_wav)
            subprocess.call(command_concat)

            # Clean up
            shutil.rmtree(tmpdir)
        else:
            # The -R to suppress dithering so that command produces identical results each time
            command = [sox_cmd, "-D", "-R", in_wav, out_wav, "speed", str(speeds)]
            if padding > 0:
                # Change speed and pad
                command += ["pad", f"{padding}s", "0s"]
            else:
                # Change speed and trim
                command += ["trim", f"{-padding}s"]
                # Note, the order of speed vs trim/pad makes a difference
                # I believe sox actually applies the speed transform after the padding.
                # but speed is so close to 1 and the padding so short that it will
                # come out about the same either way around.

            logging.info(f"Running command: {' '.join(command)}")
            subprocess.call(command)
            return

    def adjust_json_timestamps(
        self, session: str, linear_fit: Optional[Dict[str, Any]] = None
    ) -> None:
        in_json = next((self.corpus_dir / "transcriptions").rglob(f"{session}.json"))
        relative_path = in_json.relative_to(self.corpus_dir)
        out_json = self.output_dir / relative_path

        corrected_utts = []
        with open(in_json, "r") as fin, open(out_json, "w") as fout:
            data = json.load(fin)
            for segment in data:
                if "speaker" not in segment:
                    continue
                pid = segment["speaker"]
                speed = linear_fit[pid]["speed"]
                padding = linear_fit[pid]["padding"]
                delta_t = padding / 16000  # convert to seconds
                start_time = (
                    TimeFormatConverter.hms_to_seconds(
                        segment["start_time"]["original"]
                    )
                    / speed
                    + delta_t
                )
                end_time = (
                    TimeFormatConverter.hms_to_seconds(segment["end_time"]["original"])
                    / speed
                    + delta_t
                )
                segment["start_time"] = TimeFormatConverter.seconds_to_hms(start_time)
                segment["end_time"] = TimeFormatConverter.seconds_to_hms(end_time)
                corrected_utts.append(segment)

            json.dump(corrected_utts, fout, indent=2)
        return



# pylint: disable=C0415,R0913,R0914
import sys
from functools import partial
from itertools import chain
from pathlib import Path
from typing import Optional

import click
from tqdm import tqdm

from lhotse import CutSet, RecordingSet, SupervisionSet
from lhotse.bin.modes.cli_base import cli
from lhotse.parallel import ParallelExecutor
from lhotse.serialization import load_manifest_lazy_or_eager
from lhotse.utils import PythonLiteralOption, exactly_one_not_null


@cli.group()
def workflows():
    """Workflows using corpus creation tools."""
    pass


@workflows.command()
@click.argument("out_cuts", type=click.Path(allow_dash=True))
@click.option(
    "-m",
    "--recordings-manifest",
    type=click.Path(exists=True, dir_okay=False, allow_dash=True),
    help="Path to an existing recording manifest.",
)
@click.option(
    "-r",
    "--recordings-dir",
    type=click.Path(exists=True, file_okay=False),
    help="Directory with recordings. We will create a RecordingSet for it automatically.",
)
@click.option(
    "-c",
    "--cuts-manifest",
    type=click.Path(exists=True, dir_okay=False, allow_dash=True),
    help="Path to an existing cuts manifest.",
)
@click.option(
    "-e",
    "--extension",
    default="wav",
    help="Audio file extension to search for. Used with RECORDINGS_DIR.",
)
@click.option(
    "-n",
    "--model-name",
    default="base",
    help="One of Whisper variants (base, medium, large, etc.)",
)
@click.option(
    "-l",
    "--language",
    help="Language spoken in the audio. Inferred by default.",
)
@click.option(
    "-d", "--device", default="cpu", help="Device on which to run the inference."
)
@click.option("-j", "--jobs", default=1, help="Number of jobs for audio scanning.")
@click.option(
    "--force-nonoverlapping/--keep-overlapping",
    default=False,
    help="If True, the Whisper segment time-stamps will be processed to make sure they are non-overlapping.",
)
def annotate_with_whisper(
    out_cuts: str,
    recordings_manifest: Optional[str],
    recordings_dir: Optional[str],
    cuts_manifest: Optional[str],
    extension: str,
    model_name: str,
    language: Optional[str],
    device: str,
    jobs: int,
    force_nonoverlapping: bool,
):
    """
    Use OpenAI Whisper model to annotate either RECORDINGS_MANIFEST, RECORDINGS_DIR, or CUTS_MANIFEST.
    It will perform automatic segmentation, transcription, and language identification.

    RECORDINGS_MANIFEST, RECORDINGS_DIR, and CUTS_MANIFEST are mutually exclusive. If CUTS_MANIFEST
    is provided, its supervisions will be overwritten with the results of the inference.

    Note: this is an experimental feature of Lhotse, and is not guaranteed to yield
    high quality of data.
    """
    from lhotse import annotate_with_whisper as annotate_with_whisper_

    assert exactly_one_not_null(recordings_manifest, recordings_dir, cuts_manifest), (
        "Options RECORDINGS_MANIFEST, RECORDINGS_DIR, and CUTS_MANIFEST are mutually exclusive "
        "and at least one is required."
    )

    if recordings_manifest is not None:
        manifest = RecordingSet.from_file(recordings_manifest)
    elif recordings_dir is not None:
        manifest = RecordingSet.from_dir(
            recordings_dir, pattern=f"*.{extension}", num_jobs=jobs
        )
    else:
        manifest = CutSet.from_file(cuts_manifest).to_eager()

    with CutSet.open_writer(out_cuts) as writer:
        for cut in tqdm(
            annotate_with_whisper_(
                manifest,
                language=language,
                model_name=model_name,
                device=device,
                force_nonoverlapping=force_nonoverlapping,
            ),
            total=len(manifest),
            desc="Annotating with Whisper",
        ):
            writer.write(cut, flush=True)


@workflows.command()
@click.argument(
    "in_cuts", type=click.Path(exists=True, dir_okay=False, allow_dash=True)
)
@click.argument("out_cuts", type=click.Path(allow_dash=True))
@click.option(
    "-n",
    "--bundle-name",
    default="WAV2VEC2_ASR_BASE_960H",
    help="One of torchaudio pretrained 'bundle' variants (see: https://pytorch.org/audio/stable/pipelines.html)",
)
@click.option(
    "-d", "--device", default="cpu", help="Device on which to run the inference."
)
@click.option(
    "--normalize-text/--dont-normalize-text",
    default=True,
    help="By default, we'll try to normalize the text by making it uppercase and discarding symbols "
    "outside of model's character level vocabulary. If this causes issues, "
    "turn the option off and normalize the text yourself.",
)
@click.option(
    "-j",
    "--num-jobs",
    default=1,
    help="Number of parallel jobs to run.",
)
@click.option(
    "--check-language/--dont-check-language",
    default=True,
    help="If `False`, warnings about non-existent language tags in supervisions will be suppressed.",
)
def align_with_torchaudio(
    in_cuts: str,
    out_cuts: str,
    bundle_name: str,
    device: str,
    normalize_text: bool,
    num_jobs: int = 1,
    check_language: bool = True,
):
    """
    Use a pretrained ASR model from torchaudio to force align IN_CUTS (a Lhotse CutSet)
    and write the results to OUT_CUTS.
    It will attach word-level alignment information (start, end, and score) to the
    supervisions in each cut.

    This is based on a tutorial from torchaudio:
    https://pytorch.org/audio/stable/tutorials/forced_alignment_tutorial.html

    In order to use a multilingual alignment model, use `--bundle_name MMS_FA`.
    (based on the multilingual tutorial: https://pytorch.org/audio/main/tutorials/forced_alignment_for_multilingual_data_tutorial.html)

    Note: this is an experimental feature of Lhotse, and is not guaranteed to yield
    high quality of data.
    """
    from lhotse import align_with_torchaudio as align_with_torchaudio_

    cuts = load_manifest_lazy_or_eager(in_cuts)

    with CutSet.open_writer(out_cuts) as writer:
        for cut in tqdm(
            align_with_torchaudio_(
                cuts,
                bundle_name=bundle_name,
                device=device,
                normalize_text=normalize_text,
                num_jobs=num_jobs,
                verbose=False,
                check_language=check_language,
            ),
            total=len(cuts),
            desc="Aligning",
        ):
            writer.write(cut, flush=True)


@workflows.command()
@click.argument(
    "in_cuts", type=click.Path(exists=True, dir_okay=False, allow_dash=True)
)
@click.argument("out_cuts", type=click.Path(allow_dash=True))
@click.option(
    "--method",
    type=click.Choice(["independent", "conversational"]),
    default="independent",
    help="The simulation method to use: "
    "independent - each speaker is simulated independently, "
    "conversational - the speakers are simulated as a group, "
    "using overall silence/overlap statistics.",
)
# Options used with the "independent" method
@click.option(
    "--loc",
    type=float,
    default=0.0,
    help="The minimum silence duration between two consecutive utterances from the same speaker.",
    show_default=True,
)
@click.option(
    "--scale",
    type=float,
    default=2.0,
    help="The scale parameter of the exponential distribution used to sample the silence "
    "duration between two consecutive utterances from a speaker.",
    show_default=True,
)
# Options used with the "conversational" method


@click.option(
    "--same-spk-pause",
    type=float,
    default=1.0,
    help="The mean pause duration between utterances of the same speaker",
    show_default=True,
)
@click.option(
    "--diff-spk-pause",
    type=float,
    default=1.0,
    help="The mean pause duration between utterances of different speakers",
    show_default=True,
)
@click.option(
    "--diff-spk-overlap",
    type=float,
    default=2.0,
    help="The mean overlap duration between utterances of different speakers",
    show_default=True,
)
@click.option(
    "--prob-diff-spk-overlap",
    type=float,
    default=0.5,
    help="The probability of overlap between utterances of different speakers",
    show_default=True,
)
# Common options
@click.option(
    "--fit-to-supervisions",
    "-f",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Path to a supervision set to learn the distributions for simulation.",
)
@click.option(
    "--reverberate/--dont-reverberate",
    default=False,
    help="If True, the simulated meetings will be reverberated.",
)
@click.option(
    "--rir-recordings",
    "--rir",
    type=click.Path(exists=True, dir_okay=True),
    default=None,
    help="Path to a recording set containing RIRs. If provided, the simulated meetings will be "
    "reverberated using the RIRs from this set. A directory containing recording sets can also "
    "be provided, in which case each meeting will use a recording set sampled from this directory.",
)
@click.option(
    "--num-meetings",
    "-n",
    type=int,
    default=None,
    help="Number of meetings to simulate. Either this of `num_repeats` must be provided.",
)
@click.option(
    "--num-repeats",
    "-r",
    type=int,
    default=1,
    help="Number of times to repeat each input cut. The resulting cuts will be used as a finite "
    "set of utterances to use for simulation. Either this of `num_meetings` must be provided.",
)
@click.option(
    "--num-speakers-per-meeting",
    "-s",
    cls=PythonLiteralOption,
    default="2",
    help="Number of speakers per meeting. One or more integers can be provided (comma-separated). "
    "In this case, the number of speakers will be sampled uniformly from the provided list, "
    "or using the distribution provided in `speaker-count-probs`.",
)
@click.option(
    "--speaker-count-probs",
    "-p",
    cls=PythonLiteralOption,
    default=None,
    help="A list of probabilities for each speaker count. The length of the list must be "
    "equal to the number of elements in `num-speakers-per-meeting`.",
)
@click.option(
    "--max-duration-per-speaker",
    "-d",
    type=float,
    default=20.0,
    help="Maximum duration of a single speaker in a meeting.",
)
@click.option(
    "--max-utterances-per-speaker",
    "-u",
    type=int,
    default=5,
    help="Maximum number of utterances per speaker in a meeting.",
)
@click.option(
    "--allow-3fold-overlap/--no-3fold-overlap",
    default=False,
    help="If True, the simulated meetings will allow more than 2 speakers to overlap. This "
    "is only relevant for the `conversational` method.",
)
@click.option(
    "--seed",
    type=int,
    default=1234,
    help="Random seed for reproducibility.",
)
@click.option(
    "--num-jobs",
    "-j",
    type=int,
    default=1,
    help="Number of parallel jobs to run.",
)
def simulate_meetings(
    in_cuts: str,
    out_cuts: str,
    method: str,
    loc: float,
    scale: float,
    same_spk_pause: float,
    diff_spk_pause: float,
    diff_spk_overlap: float,
    prob_diff_spk_overlap: float,
    fit_to_supervisions: Optional[str],
    reverberate: bool,
    rir_recordings: Optional[str],
    **kwargs,
):
    """
    Simulate meeting-style mixtures using a provided CutSet containing single-channel
    cuts. Different simulation techniques can be selected using the `--method` option.
    Currently, the following methods are supported:

    - independent: each speaker is simulated independently, using the provided cuts as a finite
        set of utterances.

    - conversational: the speakers are simulated as a group, using overall silence/overlap
        statistics.

    The number of speakers per meeting is sampled uniformly from the range provided in
    `--num-speakers-per-meeting`.

    The number of meetings to simulate is controlled by either `--num-meetings` or
    `--num-repeats`. If the former is provided, the same number of meetings will be
    simulated. If the latter is provided, the provided cuts will be repeated `num_repeats`
    times, and the resulting cuts will be used as a finite set of utterances to use for simulation.

    The simulated meetings can be optionally reverberated using the RIRs from a provided
    recording set. If no RIRs are provided, we will use a fast random approximation technique
    to simulate the reverberation. The RIRs can be provided as a single recording set, or as
    a directory containing multiple recording sets. In the latter case, the RIRs will be sampled
    from the provided directory.

    """
    if method == "independent":
        from lhotse.workflows.meeting_simulation import (
            SpeakerIndependentMeetingSimulator,
        )

        simulator = SpeakerIndependentMeetingSimulator(loc=loc, scale=scale)

        # Remove options that are not relevant for the independent method.
        kwargs.pop("allow_3fold_overlap")
    elif method == "conversational":
        from lhotse.workflows.meeting_simulation import ConversationalMeetingSimulator

        simulator = ConversationalMeetingSimulator(
            same_spk_pause=same_spk_pause,
            diff_spk_pause=diff_spk_pause,
            diff_spk_overlap=diff_spk_overlap,
            prob_diff_spk_overlap=prob_diff_spk_overlap,
        )
    else:
        raise ValueError(f"Unknown meeting simulation method: {method}")

    if fit_to_supervisions is not None:
        print("Fitting the meeting simulator to the provided supervisions...")
        sups = load_manifest_lazy_or_eager(
            fit_to_supervisions, manifest_cls=SupervisionSet
        )
        simulator.fit(sups)

    cuts = load_manifest_lazy_or_eager(in_cuts)
    print("Simulating meetings...")
    mixed_cuts = simulator.simulate(
        cuts,
        **kwargs,
    )

    if reverberate:
        print("Reverberating the simulated meetings...")
        if rir_recordings:
            if rir_recordings.is_file():
                rirs = [
                    load_manifest_lazy_or_eager(
                        rir_recordings, manifest_cls=RecordingSet
                    )
                ]
            else:
                rirs = [
                    load_manifest_lazy_or_eager(p)
                    for p in rir_recordings.glob("*.jsonl.gz")
                ]
            mixed_cuts = simulator.reverberate(mixed_cuts, *rirs)
        else:
            mixed_cuts = simulator.reverberate(mixed_cuts)

    print("Saving the simulated meetings...")
    mixed_cuts.to_file(out_cuts)


@workflows.command()
@click.option(
    "-r",
    "--recordings-manifest",
    type=click.Path(exists=True, dir_okay=False, allow_dash=True),
    help="Path to an existing recording manifest.",
)
@click.option(
    "-o",
    "--output-supervisions-manifest",
    type=click.Path(exists=False, dir_okay=True, allow_dash=True),
    help="Path to the output supervisions manifest or a directory where it will be saved.",
)
@click.option(
    "-m",
    "--model-name",
    default="silero-vad-16k",
    help="One of activity detector: silero_vad_16k, silero_vad_8k.",
)
@click.option(
    "-d",
    "--device",
    default="cpu",
    help="Device on which to run the inference.",
)
@click.option(
    "-j",
    "--jobs",
    default=1,
    help="Number of jobs for audio scanning.",
)
@click.option(
    "--force_download",
    is_flag=True,
    help="Forced cache clearing and model downloading",
)
def activity_detection(
    recordings_manifest: str,
    output_supervisions_manifest: Optional[str],
    model_name: str,
    device: str,
    jobs: int,
    force_download: bool,
):
    """
    Use activity detection methods (e.g., Silero VAD) to detect and annotate
    the segmentation of Lhotse RecordingSets and save the results in the
    SupervisionSet manifest. The output manifest will be saved in the path
    specified by OUTPUT_SUPERVISIONS_MANIFEST. If OUTPUT_SUPERVISIONS_MANIFEST
    is not provided, the output manifest will be saved in the same directory
    as RECORDINGS_MANIFEST.

    Note: this is an experimental feature and it does not guarantee
    high-quality performance and data annotation.
    """

    import warnings

    from lhotse.workflows.activity_detection import SileroVAD8k, SileroVAD16k

    warnings.filterwarnings("ignore")

    detectors = {
        "silero_vad_8k": SileroVAD8k,
        "silero_vad_16k": SileroVAD16k,
    }
    detector_kls = detectors.get(model_name)

    if detector_kls is None:
        print(
            f"Unknown activity detector: {model_name}. "
            f"Supported detectors: {list(detectors)}"
        )
        sys.exit()

    # prepare paths and input data
    recs_path = Path(recordings_manifest).expanduser().absolute()
    if not recs_path.exists() or not recs_path.is_file():
        print(f"Recordings manifest not found: {str(recs_path)}")
        sys.exit()

    sups_path = (
        recs_path.parent
        if output_supervisions_manifest is None
        else Path(output_supervisions_manifest).expanduser().absolute()
    )
    if sups_path.is_dir():
        name = Path(recs_path).name
        for ext in [".gz", ".jsonl", ".json", ".yaml"]:
            if name.endswith(ext):  # .remove_suffix(ext) in Python 3.9
                name = name[: -len(ext)]
        name += f"_supervisions_{model_name}.jsonl.gz"
        sups_path = sups_path / name

    if not sups_path.parent.exists():
        print(f"Parent directory for output manifest does not exist: {str(sups_path)}")
        sys.exit()

    print(f"Loading recordings from {str(recordings_manifest)}...")
    recordings = RecordingSet.from_file(str(recordings_manifest))

    # run activity detection
    if force_download:  # pragma: no cover
        print("Removing model state from cache...")
        detector_kls.force_download()
    else:
        print("Checking model state in cache...")
        detector_kls("cpu")

    print(f"Making activity detection processor for {model_name!r}...")
    detector_init_fn = partial(detector_kls, device=device)
    processor = ParallelExecutor(
        init_fn=detector_init_fn,
        num_jobs=jobs,
        verbose=True,
        description="Running VAD",
    )
    print(f"Running activity detection using {model_name!r}...")
    supervisions = SupervisionSet.from_segments(
        chain.from_iterable(processor(recordings))
    )

    print(f"Saving {model_name!r} results ...")
    supervisions.to_file(str(sups_path))

    print("Results saved to:", str(sups_path), sep="\n")


@workflows.command()
@click.argument("out_cuts", type=click.Path(allow_dash=True))
@click.option(
    "-m",
    "--recordings-manifest",
    type=click.Path(exists=True, dir_okay=False, allow_dash=True),
    help="Path to an existing recording manifest.",
)
@click.option(
    "-r",
    "--recordings-dir",
    type=click.Path(exists=True, file_okay=False),
    help="Directory with recordings. We will create a RecordingSet for it automatically.",
)
@click.option(
    "-c",
    "--cuts-manifest",
    type=click.Path(exists=True, dir_okay=False, allow_dash=True),
    help="Path to an existing cuts manifest.",
)
@click.option(
    "-e",
    "--extension",
    default="wav",
    help="Audio file extension to search for. Used with RECORDINGS_DIR.",
)
@click.option(
    "-p",
    "--is-personalized-mos",
    default=False,
    help="Flag to indicate if personalized MOS score is needed or regular.",
)
@click.option("-j", "--jobs", default=1, help="Number of jobs for audio scanning.")
def annotate_dnsmos(
    out_cuts: str,
    recordings_manifest: Optional[str],
    recordings_dir: Optional[str],
    cuts_manifest: Optional[str],
    extension: str,
    is_personalized_mos: str,
    jobs: int,
):
    """
    Use Microsoft DNSMOS P.835 prediction model to annotate either RECORDINGS_MANIFEST, RECORDINGS_DIR, or CUTS_MANIFEST.
    It will predict DNSMOS P.835 score including SIG, NAK, and OVRL.

    See the original repo for more details: https://github.com/microsoft/DNS-Challenge/tree/master/DNSMOS

    RECORDINGS_MANIFEST, RECORDINGS_DIR, and CUTS_MANIFEST are mutually exclusive. If CUTS_MANIFEST
    is provided, its supervisions will be overwritten with the results of the inference.
    """
    from lhotse import annotate_dnsmos as annotate_dnsmos_

    assert exactly_one_not_null(recordings_manifest, recordings_dir, cuts_manifest), (
        "Options RECORDINGS_MANIFEST, RECORDINGS_DIR, and CUTS_MANIFEST are mutually exclusive "
        "and at least one is required."
    )

    if recordings_manifest is not None:
        manifest = RecordingSet.from_file(recordings_manifest)
    elif recordings_dir is not None:
        manifest = RecordingSet.from_dir(
            recordings_dir, pattern=f"*.{extension}", num_jobs=jobs
        )
    else:
        manifest = CutSet.from_file(cuts_manifest).to_eager()

    with CutSet.open_writer(out_cuts) as writer:
        for cut in tqdm(
            annotate_dnsmos_(
                manifest,
                is_personalized_mos=is_personalized_mos,
            ),
            total=len(manifest),
            desc="Annotating with DNSMOS P.835 prediction model",
        ):
            writer.write(cut, flush=True)



import os
import random
import types
import warnings
from contextlib import contextmanager
from functools import partial
from typing import Any, Callable, Iterable, List, Literal, Optional, TypeVar, Union

from lhotse.serialization import (
    LazyMixin,
    decode_json_line,
    deserialize_item,
    open_best,
)
from lhotse.utils import Pathlike, fastcopy, is_module_available, streaming_shuffle

T = TypeVar("T")


class AlgorithmMixin(LazyMixin, Iterable):
    """
    Helper base class with methods that are supposed to work identically
    on Lhotse manifest classes such as CutSet, RecordingSet, etc.
    """

    def filter(self, predicate: Callable[[T], bool]):
        """
        Return a new manifest containing only the items that satisfy ``predicate``.
        If the manifest is lazy, the filtering will also be applied lazily.

        :param predicate: a function that takes a cut as an argument and returns bool.
        :return: a filtered manifest.
        """
        cls = type(self)

        if self.is_lazy:
            return cls(LazyFilter(self, predicate=predicate))

        return cls.from_items(cut for cut in self if predicate(cut))

    def map(self, transform_fn: Callable[[T], T]):
        """
        Apply `transform_fn` to each item in this manifest and return a new manifest.
        If the manifest is opened lazy, the transform is also applied lazily.

        :param transform_fn: A callable (function) that accepts a single item instance
            and returns a new (or the same) instance of the same type.
            E.g. with CutSet, callable accepts ``Cut`` and returns also ``Cut``.
        :return: a new ``CutSet`` with transformed cuts.
        """
        cls = type(self)
        ans = cls(LazyMapper(self.data, fn=transform_fn))
        if self.is_lazy:
            return ans
        return ans.to_eager()

    @classmethod
    def mux(
        cls,
        *manifests,
        stop_early: bool = False,
        weights: Optional[List[Union[int, float]]] = None,
        seed: Union[int, Literal["trng", "randomized"]] = 0,
    ):
        """
        Merges multiple manifest iterables into a new iterable by lazily multiplexing them during iteration time.
        If one of the iterables is exhausted before the others, we will keep iterating until all iterables
        are exhausted. This behavior can be changed with ``stop_early`` parameter.

        :param manifests: iterables to be multiplexed.
            They can be either lazy or eager, but the resulting manifest will always be lazy.
        :param stop_early: should we stop the iteration as soon as we exhaust one of the manifests.
        :param weights: an optional weight for each iterable, affects the probability of it being sampled.
            The weights are uniform by default.
            If lengths are known, it makes sense to pass them here for uniform distribution of
            items in the expectation.
        :param seed: the random seed, ensures deterministic order across multiple iterations.
        """
        return cls(
            LazyIteratorMultiplexer(
                *manifests, stop_early=stop_early, weights=weights, seed=seed
            )
        )

    @classmethod
    def infinite_mux(
        cls,
        *manifests,
        weights: Optional[List[Union[int, float]]] = None,
        seed: Union[int, Literal["trng", "randomized"]] = 0,
        max_open_streams: Optional[int] = None,
    ):
        """
        Merges multiple manifest iterables into a new iterable by lazily multiplexing them during iteration time.
        Unlike ``mux()``, this method allows to limit the number of max open sub-iterators at any given time.

        To enable this, it performs 2-stage sampling.
        First, it samples with replacement the set of iterators ``I`` to construct a subset ``I_sub``
        of size ``max_open_streams``.
        Then, for each iteration step, it samples an iterator ``i`` from ``I_sub``,
        fetches the next item from it, and yields it.
        Once ``i`` becomes exhausted, it is replaced with a new iterator ``j`` sampled from ``I_sub``.

        .. caution:: Do not use this method with inputs that are infinitely iterable as they will
            silently break the multiplexing property by only using a subset of the input iterables.

        .. caution:: This method is not recommended for multiplexing for a small amount of iterations,
            as it may be much less accurate than ``mux()`` depending on the number of open streams,
            iterable sizes, and the random seed.

        :param manifests: iterables to be multiplexed.
            They can be either lazy or eager, but the resulting manifest will always be lazy.
        :param weights: an optional weight for each iterable, affects the probability of it being sampled.
            The weights are uniform by default.
            If lengths are known, it makes sense to pass them here for uniform distribution of
            items in the expectation.
        :param seed: the random seed, ensures deterministic order across multiple iterations.
        :param max_open_streams: the number of iterables that can be open simultaneously at any given time.
        """
        return cls(
            LazyInfiniteApproximateMultiplexer(
                *manifests,
                weights=weights,
                seed=seed,
                max_open_streams=max_open_streams,
            )
        )

    def shuffle(
        self,
        rng: Optional[random.Random] = None,
        buffer_size: int = 10000,
    ):
        """
        Shuffles the elements and returns a shuffled variant of self.
        If the manifest is opened lazily, performs shuffling on-the-fly with a fixed buffer size.

        :param rng: an optional instance of ``random.Random`` for precise control of randomness.
        :return: a shuffled copy of self, or a manifest that is shuffled lazily.
        """
        cls = type(self)

        if rng is None:
            rng = random

        if self.is_lazy:
            return cls(LazyShuffler(self.data, buffer_size=buffer_size, rng=rng))
        else:
            new: List = self.data.copy()
            rng.shuffle(new)
            return cls(new)

    def repeat(self, times: Optional[int] = None, preserve_id: bool = False):
        """
        Return a new, lazily evaluated manifest that iterates over the original elements ``times``
        number of times.

        :param times: how many times to repeat (infinite by default).
        :param preserve_id: when ``True``, we won't update the element ID with repeat number.
        :return: a repeated manifest.
        """
        cls = type(self)
        return cls(LazyRepeater(self, times=times, preserve_id=preserve_id))

    def __add__(self, other):
        cls = type(self)
        return cls(LazyIteratorChain(self.data, other.data))


class Dillable:
    """
    Mix-in that will leverage ``dill`` instead of ``pickle``
    when pickling an object.

    It is useful when the user can't avoid ``pickle`` (e.g. in multiprocessing),
    but needs to use unpicklable objects such as lambdas.

    If ``dill`` is not installed, it defers to what ``pickle`` does by default.
    """

    _ENABLED_VALUES = {"1", "True", "true", "yes"}

    def __getstate__(self):
        if is_dill_enabled():
            import dill

            return dill.dumps(self.__dict__)
        else:
            return self.__dict__

    def __setstate__(self, state):
        if is_dill_enabled():
            import dill

            self.__dict__ = dill.loads(state)
        else:
            self.__dict__ = state


def is_dill_enabled(_ENABLED_VALUES=frozenset(("1", "True", "true", "yes"))) -> bool:
    """Returns bool indicating if dill-based pickling in Lhotse is enabled or not."""
    return (
        is_module_available("dill")
        and os.environ.get("LHOTSE_DILL_ENABLED", "0") in _ENABLED_VALUES
    )


def set_dill_enabled(value: bool) -> None:
    """Enable or disable dill-based pickling in Lhotse."""
    assert is_module_available("dill"), (
        "Cannot enable dill because dill is not installed. "
        "Please run 'pip install dill' and try again."
    )
    # We use os.environ here so that sub-processes / forks will inherit this value
    os.environ["LHOTSE_DILL_ENABLED"] = "1" if value else "0"


@contextmanager
def dill_enabled(value: bool):
    """
    Context manager that overrides the setting of Lhotse's dill-backed pickling
    and restores the previous value after exit.

    Example::

        >>> import pickle
        ... with dill_enabled(True):
        ...    pickle.dump(CutSet(...).filter(lambda c: c.duration < 5), open("cutset.pickle", "wb"))
    """
    previous = is_dill_enabled()
    set_dill_enabled(value)
    yield
    set_dill_enabled(previous)


class LazyTxtIterator:
    """
    LazyTxtIterator is a thin wrapper over builtin ``open`` function to
    iterate over lines in a (possibly compressed) text file.
    It can also provide the number of lines via __len__ via fast newlines counting.
    """

    def __init__(self, path: Pathlike, as_text_example: bool = True) -> None:
        self.path = path
        self.as_text_example = as_text_example
        self._len = None

    def __iter__(self):
        from lhotse.cut.text import TextExample

        tot = 0
        with open_best(self.path, "r") as f:
            for line in f:
                line = line.strip()
                if self.as_text_example:
                    line = TextExample(line)
                yield line
                tot += 1
        if self._len is None:
            self._len = tot

    def __len__(self) -> int:
        if self._len is None:
            self._len = count_newlines_fast(self.path)
        return self._len


class LazyJsonlIterator:
    """
    LazyJsonlIterator provides the ability to read JSON lines as Python dicts.
    It can also provide the number of lines via __len__ via fast newlines counting.
    """

    def __init__(self, path: Pathlike) -> None:
        self.path = path
        self._len = None

    def __iter__(self):
        tot = 0
        with open_best(self.path, "r") as f:
            for line in f:
                data = decode_json_line(line)
                yield data
                tot += 1
        if self._len is None:
            self._len = tot

    def __len__(self) -> int:
        if self._len is None:
            self._len = count_newlines_fast(self.path)
        return self._len


class LazyManifestIterator(Dillable):
    """
    LazyManifestIterator provides the ability to read Lhotse objects from a
    JSONL file on-the-fly, without reading its full contents into memory.

    This class is designed to be a partial "drop-in" replacement for ordinary dicts
    to support lazy loading of RecordingSet, SupervisionSet and CutSet.
    Since it does not support random access reads, some methods of these classes
    might not work properly.
    """

    def __init__(self, path: Pathlike) -> None:
        self.source = LazyJsonlIterator(path)

    @property
    def path(self) -> Pathlike:
        return self.source.path

    def __iter__(self):
        yield from map(deserialize_item, self.source)

    def __len__(self) -> int:
        return len(self.source)

    def __add__(self, other) -> "LazyIteratorChain":
        return LazyIteratorChain(self, other)


class LazyIteratorChain(Dillable):
    """
    A thin wrapper over multiple iterators that enables to combine lazy manifests
    in Lhotse. It iterates all underlying iterables sequentially.

    It also supports shuffling the sub-iterators when it's iterated over.
    This can be used to implement sharding (where each iterator is a shard)
    with randomized shard order. Every iteration of this object will increment
    an internal counter so that the next time it's iterated, the order of shards
    is again randomized.

    .. note:: if any of the input iterables is a dict, we'll iterate only its values.
    """

    def __init__(
        self,
        *iterators: Iterable,
        shuffle_iters: bool = False,
        seed: Optional[Union[int, Literal["trng", "randomized"]]] = None,
    ) -> None:
        self.iterators = []
        self.shuffle_iters = shuffle_iters
        self.seed = seed
        self.num_iters = 0
        for it in iterators:
            # Auto-flatten LazyIteratorChain instances if any are passed
            if isinstance(it, LazyIteratorChain):
                for sub_it in it.iterators:
                    self.iterators.append(sub_it)
            else:
                self.iterators.append(it)

    def __iter__(self):
        from lhotse.dataset.dataloading import resolve_seed

        iterators = self.iterators
        if self.shuffle_iters:
            if self.seed is None:
                rng = random  # global Python RNG
            else:
                rng = random.Random(resolve_seed(self.seed) + self.num_iters)
            rng.shuffle(iterators)
            self.num_iters += 1
        for it in iterators:
            if isinstance(it, dict):
                it = it.values()
            yield from it

    def __len__(self) -> int:
        return sum(len(it) for it in self.iterators)

    def __add__(self, other) -> "LazyIteratorChain":
        return LazyIteratorChain(self, other)


class LazyIteratorMultiplexer(Dillable):
    """
    A wrapper over multiple iterators that enables to combine lazy manifests in Lhotse.
    During iteration, unlike :class:`.LazyIteratorChain`, :class:`.LazyIteratorMultiplexer`
    at each step randomly selects the iterable used to yield an item.

    Since the iterables might be of different length, we provide a ``weights`` parameter
    to let the user decide which iterables should be sampled more frequently than others.
    When an iterable is exhausted, we will keep sampling from the other iterables, until
    we exhaust them all, unless ``stop_early`` is set to ``True``.
    """

    def __init__(
        self,
        *iterators: Iterable,
        stop_early: bool = False,
        weights: Optional[List[Union[int, float]]] = None,
        seed: Union[int, Literal["trng", "randomized"]] = 0,
    ) -> None:
        self.iterators = list(iterators)
        self.stop_early = stop_early
        self.seed = seed

        assert (
            len(self.iterators) > 1
        ), "There have to be at least two iterables to multiplex."

        if weights is None:
            self.weights = [1] * len(self.iterators)
        else:
            self.weights = weights

        assert len(self.iterators) == len(self.weights)

    def __iter__(self):
        from lhotse.dataset.dataloading import resolve_seed

        rng = random.Random(resolve_seed(self.seed))
        iters = [iter(it) for it in self.iterators]
        exhausted = [False for _ in range(len(iters))]

        def should_continue():
            if self.stop_early:
                return not any(exhausted)
            else:
                return not all(exhausted)

        while should_continue():
            active_indexes, active_weights = zip(
                *[
                    (i, w)
                    for i, (is_exhausted, w) in enumerate(zip(exhausted, self.weights))
                    if not is_exhausted
                ]
            )
            idx = rng.choices(active_indexes, weights=active_weights, k=1)[0]
            selected = iters[idx]
            try:
                item = next(selected)
                yield item
            except StopIteration:
                exhausted[idx] = True
                continue

    def __len__(self) -> int:
        return sum(len(it) for it in self.iterators)

    def __add__(self, other) -> "LazyIteratorChain":
        return LazyIteratorChain(self, other)


class LazyInfiniteApproximateMultiplexer(Dillable):
    """
    A variant of :class:`.LazyIteratorMultiplexer` that allows to control the number of
    iterables that are simultaneously open.

    It is useful for large-scale data sets where opening multiple file handles in
    many processes leads to exhaustion of the operating system resources.

    If the data sets are sharded, it is recommended to pass each shard as a separate iterator
    when creating objects of this class. It is OK to assign a dataset-level weight to each shard
    (e.g., if a dataset has a weight of 0.5, assign weight 0.5 to each of its shards).

    There are several differences between this class and :class:`.LazyIteratorMultiplexer`:
    * Objects of this class are infinite iterators.
    * We hold a list of ``max_open_streams`` open iterators at any given time.
        This list is filled by sampling input iterators with replacement.

    These differences are necessary to guarantee the weighted sampling property.
    If we did not sample with replacement or make it infinite, we would simply
    exhaust highly-weighted iterators towards the beginning of each "epoch"
    and keep sampling only lowly-weighted iterators towards the end of each "epoch".
    """

    def __init__(
        self,
        *iterators: Iterable,
        stop_early: bool = False,
        weights: Optional[List[Union[int, float]]] = None,
        seed: Union[int, Literal["trng", "randomized"]] = 0,
        max_open_streams: Optional[int] = None,
    ) -> None:
        self.iterators = list(iterators)
        self.stop_early = stop_early
        self.seed = seed
        self.max_open_streams = max_open_streams
        if max_open_streams is None or max_open_streams > len(self.iterators):
            self.max_open_streams = len(self.iterators)

        assert len(self.iterators) > 0
        self.weights = weights
        if weights is None:
            self.weights = [1] * len(self.iterators)
        assert len(self.iterators) == len(self.weights)
        assert (
            self.max_open_streams is None or self.max_open_streams >= 1
        ), f"{self.max_open_streams=}"

    def __iter__(self):
        """
        Assumptions
        - we have N streams but can only open M at the time (M < N)
        - the streams are finite
        - each stream needs to be "short" to ensure the mux property
        - each stream may be interpreted as a shard belonging to some larger group of streams
          (e.g. multiple shards of a given dataset).
        """
        from lhotse.dataset.dataloading import resolve_seed

        rng = random.Random(resolve_seed(self.seed))

        def shuffled_streams():
            # Create an infinite iterable of our streams.
            # Assume N is "small" enough that shuffling it will be quick
            #
            # we need to incorporate weights into shuffling here
            # and sample iterators with replacement.
            # consider it0=[shard00, shard01] with weight 0.95
            # and      it1=[shard10, shard11] with weight 0.05
            # so we have 4 streams [shard{01}{01}]
            # if we just shuffle randomly and sample without replacement
            # per each "epoch" (epoch = 4 shards) then we would have
            # ignored the weights because we'll just exhaust it0 shards
            # towards the beginning of an "epoch" and then keep yielding
            # from it1 shards until the epoch is finished and we can sample
            # from it0 again...
            indexes = list(range(len(self.iterators)))
            while True:
                selected = rng.choices(indexes, self.weights, k=1)[0]
                yield self.iterators[selected], self.weights[selected]

        # Initialize an infinite sequence of finite streams.
        # It is sampled with weights and replacement from ``self.iterators``,
        # which are of length N.
        stream_source = shuffled_streams()

        # Sample the first M active streams to be multiplexed.
        # As streams get depleted, we will replace them with
        # new streams sampled from the stream source.
        active_streams = [None] * self.max_open_streams
        active_weights = [None] * self.max_open_streams
        stream_indexes = list(range(self.max_open_streams))

        def sample_new_stream_at(pos: int) -> None:
            sampled_stream, sampled_weight = next(stream_source)
            active_streams[pos] = iter(sampled_stream)
            active_weights[pos] = sampled_weight

        for stream_pos in range(self.max_open_streams):
            sample_new_stream_at(stream_pos)

        # The actual multiplexing loop.
        while True:
            # Select a stream from the currently active streams.
            # We actually sample an index so that we know which position
            # to replace if a stream is exhausted.
            stream_pos = rng.choices(
                stream_indexes,
                weights=active_weights if sum(active_weights) > 0 else None,
                k=1,
            )[0]
            selected = active_streams[stream_pos]
            try:
                # Sample from the selected stream.
                item = next(selected)
                yield item
            except StopIteration:
                # The selected stream is exhausted. Replace it with another one,
                # and return a sample from the newly opened stream.
                sample_new_stream_at(stream_pos)
                item = next(active_streams[stream_pos])
                yield item


class LazyShuffler(Dillable):
    """
    A wrapper over an iterable that enables lazy shuffling.
    The shuffling algorithm is reservoir-sampling based.
    See :func:`lhotse.utils.streaming_shuffle` for details.
    """

    def __init__(
        self,
        iterator: Iterable,
        buffer_size: int = 10000,
        rng: Optional[random.Random] = None,
    ) -> None:
        self.iterator = iterator
        self.buffer_size = buffer_size
        self.rng = rng

    def __iter__(self):
        return iter(
            streaming_shuffle(
                iter(self.iterator),
                bufsize=self.buffer_size,
                rng=self.rng,
            )
        )

    def __len__(self) -> int:
        return len(self.iterator)

    def __add__(self, other) -> "LazyIteratorChain":
        return LazyIteratorChain(self, other)


class LazyFilter(Dillable):
    """
    A wrapper over an iterable that enables lazy filtering.
    It works like Python's `filter` built-in by applying the filter predicate
    to each element and yielding it further if predicate returned ``True``.
    """

    def __init__(self, iterator: Iterable, predicate: Callable[[Any], bool]) -> None:
        self.iterator = iterator
        self.predicate = predicate
        assert callable(
            self.predicate
        ), f"LazyFilter: 'predicate' arg must be callable (got {predicate})."
        if (
            isinstance(self.predicate, types.LambdaType)
            and self.predicate.__name__ == "<lambda>"
            and not is_module_available("dill")
        ):
            warnings.warn(
                "A lambda was passed to LazyFilter: it may prevent you from forking this process. "
                "If you experience issues with num_workers > 0 in torch.utils.data.DataLoader, "
                "try passing a regular function instead."
            )

    def __iter__(self):
        return filter(self.predicate, self.iterator)

    def __add__(self, other) -> "LazyIteratorChain":
        return LazyIteratorChain(self, other)

    def __len__(self) -> int:
        raise TypeError(
            "LazyFilter does not support __len__ because it would require "
            "iterating over the whole iterator, which is not possible in a lazy fashion. "
            "If you really need to know the length, convert to eager mode first using "
            "`.to_eager()`. Note that this will require loading the whole iterator into memory."
        )


class LazyMapper(Dillable):
    """
    A wrapper over an iterable that enables lazy function evaluation on each item.
    It works like Python's `map` built-in by applying a callable ``fn``
    to each element ``x`` and yielding the result of ``fn(x)`` further.

    New in Lhotse v1.22.0: ``apply_fn`` can be provided to decide whether ``fn`` should be applied
        to a given example or not (in which case it will return it as-is, i.e., it does not filter).
    """

    def __init__(
        self,
        iterator: Iterable,
        fn: Callable[[Any], Any],
        apply_fn: Optional[Callable[[Any], bool]] = None,
    ) -> None:
        self.iterator = iterator
        self.fn = fn
        self.apply_fn = apply_fn
        assert callable(self.fn), f"LazyMapper: 'fn' arg must be callable (got {fn})."
        if self.apply_fn is not None:
            assert callable(
                self.apply_fn
            ), f"LazyMapper: 'apply_fn' arg must be callable (got {fn})."
        if (
            (isinstance(self.fn, types.LambdaType) and self.fn.__name__ == "<lambda>")
            or (
                isinstance(self.apply_fn, types.LambdaType)
                and self.apply_fn.__name__ == "<lambda>"
            )
            and not is_dill_enabled()
        ):
            warnings.warn(
                "A lambda was passed to LazyMapper: it may prevent you from forking this process. "
                "If you experience issues with num_workers > 0 in torch.utils.data.DataLoader, "
                "try passing a regular function instead."
            )

    def __iter__(self):
        if self.apply_fn is None:
            yield from map(self.fn, self.iterator)
        else:
            for item in self.iterator:
                if self.apply_fn(item):
                    ans = self.fn(item)
                else:
                    ans = item
                yield ans

    def __len__(self) -> int:
        return len(self.iterator)

    def __add__(self, other) -> "LazyIteratorChain":
        return LazyIteratorChain(self, other)


class LazyFlattener(Dillable):
    """
    A wrapper over an iterable of collections that flattens it to an iterable of items.

    Example::

        >>> list_of_cut_sets: List[CutSet] = [CutSet(...), CutSet(...)]
        >>> list_of_cuts: List[Cut] = list(LazyFlattener(list_of_cut_sets))
    """

    def __init__(self, iterator: Iterable) -> None:
        self.iterator = iterator

    def __iter__(self):
        for cuts in self.iterator:
            yield from cuts

    def __add__(self, other) -> "LazyIteratorChain":
        return LazyIteratorChain(self, other)

    def __len__(self) -> int:
        raise TypeError(
            "LazyFlattener does not support __len__ because it would require "
            "iterating over the whole iterator, which is not possible in a lazy fashion. "
            "If you really need to know the length, convert to eager mode first using "
            "`.to_eager()`. Note that this will require loading the whole iterator into memory."
        )


class LazyRepeater(Dillable):
    """
    A wrapper over an iterable that enables to repeat it N times or infinitely (default).
    """

    def __init__(
        self, iterator: Iterable, times: Optional[int] = None, preserve_id: bool = False
    ) -> None:
        self.iterator = iterator
        self.times = times
        self.preserve_id = preserve_id
        assert self.times is None or self.times > 0

    def __iter__(self):
        epoch = 0
        while self.times is None or epoch < self.times:
            if self.preserve_id:
                iterator = self.iterator
            else:
                iterator = LazyMapper(
                    self.iterator, partial(attach_repeat_idx_to_id, idx=epoch)
                )
            at_least_once = False
            for item in iterator:
                at_least_once = True
                yield item
            if not at_least_once:
                return  # Detect empty iterables to avoid hanging the program.
            epoch += 1

    def __len__(self) -> int:
        if self.times is None:
            raise TypeError(
                f"object of type '{type(self).__name__}' is an infinite iterator"
            )
        return len(self.iterator) * self.times

    def __add__(self, other) -> "LazyIteratorChain":
        return LazyIteratorChain(self, other)


class LazySlicer(Dillable):
    """
    A wrapper over an iterable that enables selecting k-th element every n elements.
    """

    def __init__(self, iterator: Iterable, k: int, n: int) -> None:
        self.iterator = iterator
        assert (
            k < n
        ), f"When selecting k-th element every n elements, k must be less than n (got k={k} n={n})."
        self.k = k
        self.n = n

    def __iter__(self):
        for idx, item in enumerate(self.iterator):
            if idx % self.n == self.k:
                yield item

    def __add__(self, other) -> "LazyIteratorChain":
        return LazyIteratorChain(self, other)

    def __len__(self) -> int:
        raise TypeError(
            "LazySlicer does not support __len__ because it would require "
            "iterating over the whole iterator, which is not possible in a lazy fashion. "
            "If you really need to know the length, convert to eager mode first using "
            "`.to_eager()`. Note that this will require loading the whole iterator into memory."
        )


def attach_repeat_idx_to_id(item: Any, idx: int) -> Any:
    if not hasattr(item, "id"):
        return item
    return fastcopy(item, id=f"{item.id}_repeat{idx}")


def count_newlines_fast(path: Pathlike):
    """
    Counts newlines in a file using buffered chunk reads.
    The fastest possible option in Python according to:
    https://stackoverflow.com/a/68385697/5285891
    (This is a slightly modified variant of that answer.)
    """

    def _make_gen(reader):
        b = reader(2**16)
        while b:
            yield b
            b = reader(2**16)

    read_mode = "rb" if not str(path) == "-" else "r"
    with open_best(path, read_mode) as f:
        count = sum(buf.count(b"\n") for buf in _make_gen(f.read))
    return count



"""
Description taken from the abstract of paper:
"GigaSpeech 2: An Evolving, Large-Scale and Multi-domain ASR Corpus for Low-Resource Languages with Automated Crawling, Transcription and Refinement"
https://arxiv.org/abs/2406.11546
"""

import logging
from collections import defaultdict
from concurrent.futures.process import ProcessPoolExecutor
from pathlib import Path
from typing import Dict, Optional, Sequence, Tuple, Union

from tqdm.auto import tqdm

from lhotse import load_manifest
from lhotse.audio import Recording, RecordingSet
from lhotse.qa import fix_manifests, validate_recordings_and_supervisions
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike

GIGASPEECH2_URL = "https://huggingface.co/datasets/speechcolab/gigaspeech2"

GIGASPEECH2_LANGS = ("th", "id", "vi")
GIGASPEECH2_SPLITS = ("train_raw", "train_refined", "dev", "test")


def _read_manifests_if_cached(
    output_dir: Optional[Pathlike],
    language: str,
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Returns:
        {
          "train_raw": {"recordings": ..., "supervisions": ...},
          "train_refined": ...,
          "dev": ...,
          "test": ...,
        }
    """
    if output_dir is None:
        return {}
    manifests = defaultdict(dict)
    for part in ["train_raw", "train_refined", "dev", "test"]:
        for manifest in ["recordings", "supervisions"]:
            path = output_dir / f"gigaspeech2-{language}_{manifest}_{part}.jsonl.gz"
            if not path.is_file():
                continue
            manifests[part][manifest] = load_manifest(path)
    return manifests


def _parse_utterance(
    lang: str,
    part_dir: Pathlike,
    audio_info: Pathlike,
) -> Optional[Tuple[Recording, SupervisionSegment]]:
    segment_id, text = audio_info.split("\t")
    audio_path = part_dir.joinpath(*segment_id.split("-")[:-1]) / f"{segment_id}.wav"
    audio_path = audio_path.resolve()

    if not audio_path.is_file():
        logging.warning(f"No such file: {audio_path}")
        return None

    recording = Recording.from_file(
        path=audio_path,
        recording_id=segment_id,
    )

    segment = SupervisionSegment(
        id=segment_id,
        recording_id=segment_id,
        start=0.0,
        duration=recording.duration,
        channel=0,
        language=lang,
        text=text.strip(),
    )

    return recording, segment


def prepare_gigaspeech2(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
    languages: Union[str, Sequence[str]] = "auto",
    num_jobs: int = 1,
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions
    :param corpus_dir: Path to the GigaSpeech 2 dataset.
    :param output_dir: Pathlike, the path where to write the manifests.
    :param languages: 'auto' (prepare all discovered data) or a list of language codes.
    :return: a Dict whose key is the dataset part, and the value is Dicts with the keys 'recordings' and 'supervisions'.
    """
    corpus_dir = Path(corpus_dir)
    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"
    corpus_dir = corpus_dir / "data"

    if languages == "auto":
        languages = set(GIGASPEECH2_LANGS).intersection(
            path.name for path in corpus_dir.glob("*")
        )
        if not languages:
            raise ValueError(
                f"Could not find any of GigaSpeech 2 languages in: {corpus_dir}"
            )
    elif isinstance(languages, str):
        languages = [languages]

    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

    manifests = defaultdict(dict)

    for lang in tqdm(languages, desc="Processing GigaSpeech 2 languages"):
        logging.info(f"Language: {lang}")
        lang_dir = corpus_dir / lang

        # Maybe the manifests already exist: we can read them and save a bit of preparation time.
        lang_manifests = _read_manifests_if_cached(output_dir=output_dir, language=lang)

        for part in tqdm(GIGASPEECH2_SPLITS, desc="Processing GigaSpeech 2 subset"):
            logging.info(f"Processing GigaSpeech 2 subset: {part}")
            if part in lang_manifests:
                logging.info(f"GigaSpeech 2 {lang} {part} already prepared - skipping.")
                continue

            with RecordingSet.open_writer(
                output_dir / f"gigaspeech2-{lang}_recordings_{part}.jsonl.gz"
            ) as rec_writer, SupervisionSet.open_writer(
                output_dir / f"gigaspeech2-{lang}_supervisions_{part}.jsonl.gz"
            ) as sup_writer:
                part_dir = lang_dir / part.replace("_raw", "").replace("_refined", "")
                tsv_path = lang_dir / f"{part}.tsv"

                audio_infos = []
                with open(tsv_path) as f:
                    audio_infos = f.read().splitlines()

                with ProcessPoolExecutor(num_jobs) as ex:
                    futures = []
                    for audio_info in tqdm(audio_infos, desc="Distributing tasks"):
                        futures.append(
                            ex.submit(_parse_utterance, lang, part_dir, audio_info)
                        )

                    for future in tqdm(futures, desc="Processing"):
                        result = future.result()
                        if result is None:
                            continue
                        recording, segment = result

                        # Fix and validate the recording + supervisions
                        recordings, segments = fix_manifests(
                            recordings=RecordingSet.from_recordings([recording]),
                            supervisions=SupervisionSet.from_segments([segment]),
                        )
                        validate_recordings_and_supervisions(
                            recordings=recordings, supervisions=segments
                        )

                        # Write the manifests
                        rec_writer.write(recordings[0])
                        sup_writer.write(segments[0])

                lang_manifests[part] = {
                    "recordings": RecordingSet.from_jsonl_lazy(rec_writer.path),
                    "supervisions": SupervisionSet.from_jsonl_lazy(sup_writer.path),
                }

        manifests[lang] = lang_manifests

    return dict(manifests)



"""
KsponSpeech is a large-scale spontaneous speech corpus of Korean.
This corpus contains 969 hours of open-domain dialogue utterances,
spoken by about 2,000 native Korean speakers in a clean environment.

All data were constructed by recording the dialogue between two people
freely conversing on a variety of topics and manually transcribing the utterances.

The transcription provides a dual transcription consisting of orthography and pronunciation,
and disfluency tags for the spontaneity of speech, such as filler words, repeated words, and word fragments.

The original audio data has a PCM extension.
During preprocessing, it is converted into a file in the FLAC extension and saved anew.

KsponSpeech is publicly available on an open data hub site of the Korea government.
The dataset must be downloaded manually.

For more details, please visit:

Dataset: https://aihub.or.kr/aihubdata/data/view.do?currMenu=115&topMenu=100&aihubDataSe=realm&dataSetSn=123
Paper: https://www.mdpi.com/2076-3417/10/19/6936
"""

import logging
import re
from concurrent.futures.thread import ThreadPoolExecutor
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Tuple, Union

import soundfile as sf
from tqdm.auto import tqdm

from lhotse import fix_manifests, validate_recordings_and_supervisions
from lhotse.audio import Recording, RecordingSet
from lhotse.recipes.utils import manifests_exist, read_manifests_if_cached
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike

KSPONSPEECH = (
    "train",
    "dev",
    "eval_clean",
    "eval_other",
)


def normalize(
    raw_content: str,
    normalize_text: str = "default",
) -> Tuple[str, str]:
    """
    Normalizing KsponSpeech text datasets with '.trn' extension.
    Perform the following processing.

    1. Separate file name and text labeling from raw content using separator ' :: ';
    2. Remove noise labeling characters (e.g. `o/`, `b/`...);
    3. Remove the actual pronunciation from the text labeling; use the spelling content;
    4. Remove other special characters and double spaces from text labeling.

    :param raw_content: a raw text labeling content containing file name and text labeling.
    :param normalize_text: str, the text normalization type, "default" or "none".
    :return: a tuple with file name and normalized text labeling.
    """
    if len(raw_content) == 0:
        return ""

    original_content_id, content = raw_content.split(" :: ")

    if normalize_text == "none":
        return original_content_id, content

    elif normalize_text == "default":
        content = re.sub(r"[a-z]/", "", content)
        content = re.sub(r"\((.*?)\)/\((.*?)\)", r"\1", content)
        content = content.replace("*", "")
        content = content.replace("+", "")
        content = content.replace("/", "")
        content = re.sub(r"\s+", " ", content)

        return original_content_id, content.strip()


def prepare_ksponspeech(
    corpus_dir: Pathlike,
    dataset_parts: Union[str, Sequence[str]] = "all",
    output_dir: Optional[Pathlike] = None,
    num_jobs: int = 1,
    normalize_text: str = "default",
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions.
    When all the manifests are available in the ``output_dir``, it will simply read and return them.

    :param corpus_dir: Pathlike, the path of the data dir.
    :param dataset_parts: string or sequence of strings representing dataset part names, e.g. 'train', 'dev'.
        By default, we will infer all parts.
    :param output_dir: Pathlike, the path where to write the manifests.
    :param num_jobs: int, number of parallel threads used for 'parse_utterance' calls.
    :param normalize_text: str, the text normalization type, "default" or "none".
    :return: a Dict whose key is the dataset part, and the value is Dicts with the keys 'audio' and 'supervisions'.
    """
    corpus_dir = Path(corpus_dir)
    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"

    if dataset_parts == "all":
        dataset_parts = set(KSPONSPEECH)

    elif isinstance(dataset_parts, str):
        dataset_parts = [dataset_parts]

    manifests = {}

    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)
        # Maybe the manifests already exist: we can read them and save a bit of preparation time.
        manifests = read_manifests_if_cached(
            dataset_parts=dataset_parts,
            output_dir=output_dir,
            prefix="ksponspeech",
            suffix="jsonl.gz",
            lazy=True,
        )

    with ThreadPoolExecutor(num_jobs) as ex:
        for part in tqdm(dataset_parts, desc="Dataset parts"):
            logging.info(f"Processing KsponSpeech subset: {part}")
            if manifests_exist(
                part=part,
                output_dir=output_dir,
                prefix="ksponspeech",
                suffix="jsonl.gz",
            ):
                logging.info(f"KsponSpeech subset: {part} already prepared - skipping.")
                continue

            recordings = []
            supervisions = []
            futures = []

            trans_path = corpus_dir / f"{part}.trn"
            with open(trans_path) as f:
                for line in f:
                    futures.append(
                        ex.submit(
                            parse_utterance, corpus_dir, part, line, normalize_text
                        )
                    )

            for future in tqdm(futures, desc="Processing", leave=False):
                result = future.result()
                if result is None:
                    continue
                recording, segment = result
                recordings.append(recording)
                supervisions.append(segment)

            recording_set = RecordingSet.from_recordings(recordings)
            supervision_set = SupervisionSet.from_segments(supervisions)

            recording_set, supervision_set = fix_manifests(
                recording_set, supervision_set
            )
            validate_recordings_and_supervisions(recording_set, supervision_set)

            if output_dir is not None:
                supervision_set.to_file(
                    output_dir / f"ksponspeech_supervisions_{part}.jsonl.gz"
                )
                recording_set.to_file(
                    output_dir / f"ksponspeech_recordings_{part}.jsonl.gz"
                )

            manifests[part] = {
                "recordings": recording_set,
                "supervisions": supervision_set,
            }

    return manifests


def pcm_to_flac(
    pcm_path: Union[str, Path],
    flac_path: Union[str, Path],
    sample_rate: Optional[int] = 16000,
    channels: Optional[int] = 1,
    bit_depth: Optional[int] = 16,
) -> Path:
    # typecasting
    pcm_path = Path(pcm_path)
    flac_path = Path(flac_path)

    data, _ = sf.read(
        pcm_path,
        channels=channels,
        samplerate=sample_rate,
        format="RAW",
        subtype="PCM_16",
    )

    sf.write(flac_path, data, sample_rate, format="FLAC")
    return flac_path


def parse_utterance(
    corpus_dir: Pathlike,
    part: str,
    line: str,
    normalize_text: str = "default",
) -> Optional[Tuple[Recording, SupervisionSegment]]:
    corpus_dir = Path(corpus_dir)
    audio_path, normalized_line = normalize(line, normalize_text)
    if "eval" in part:
        audio_path = audio_path.split("/", maxsplit=1)[1]

    audio_path = corpus_dir / audio_path
    recording_id = audio_path.stem

    # Create the Recording first
    if not audio_path.is_file():
        logging.warning(f"No such file: {audio_path}")
        return None
    flac_path = audio_path.with_suffix(".flac")
    flac_path = pcm_to_flac(audio_path, flac_path)
    recording = Recording.from_file(flac_path, recording_id=recording_id)
    # Then, create the corresponding supervisions
    segment = SupervisionSegment(
        id=recording_id,
        recording_id=recording_id,
        start=0.0,
        duration=recording.duration,
        channel=0,
        language="Korean",
        text=normalized_line,
    )
    return recording, segment



from typing import Sequence, Union

import click

from lhotse.bin.modes import prepare
from lhotse.recipes.csj import prepare_csj
from lhotse.utils import Pathlike

__all__ = ["csj"]


@prepare.command(context_settings=dict(show_default=True))
@click.argument("corpus_dir", type=click.Path(exists=True, dir_okay=True))
@click.argument("manifest_dir", type=click.Path())
@click.option(
    "-t",
    "--transcript-dir",
    type=click.Path(),
    default=None,
    help=(
        "Directory to save parsed transcripts in txt format, with "
        "valid and eval sets created from the core and noncore datasets. "
        "If not provided, this script will not create valid and eval "
        "sets."
    ),
)
@click.option(
    "-p",
    "--dataset-parts",
    type=str,
    default=None,
    multiple=True,
    help=(
        "List of dataset parts to prepare. "
        "To prepare multiple parts, pass each with `-p` "
        "Example: `-p eval1 -p eval2`"
    ),
)
@click.option(
    "-j",
    "--num-jobs",
    type=int,
    default=1,
    help="How many threads to use (can give good speed-ups with slow disks).",
)
def csj(
    corpus_dir: Pathlike,
    manifest_dir: Pathlike,
    dataset_parts: Union[str, Sequence[str]],
    transcript_dir: Pathlike,
    num_jobs: int,
):
    "Prepare Corpus of Spontaneous Japanese"

    prepare_csj(
        corpus_dir=corpus_dir,
        manifest_dir=manifest_dir,
        dataset_parts=dataset_parts,
        transcript_dir=transcript_dir,
        nj=num_jobs,
    )



from dataclasses import dataclass
from typing import Optional

import numpy as np

from lhotse.augmentation.transform import AudioTransform


@dataclass
class Clipping(AudioTransform):
    """
    Apply clipping to audio signal.

    This augmentation simulates the effect of clipping that occurs when audio levels exceed
    the dynamic range of recording equipment or when audio is amplified beyond its limits.

    Clips input signal to [-1, 1] range.

    :param gain_db: The amount of gain in decibels to apply before clipping (and to revert back to original level after).
    :param hard: If True, apply hard clipping (sharp cutoff); otherwise, apply soft clipping (saturation).
    :param normalize: If True, normalize the input signal to 0 dBFS before applying clipping.
    """

    hard: bool = False
    gain_db: float = 0.0
    normalize: bool = True

    def __call__(
        self,
        samples: np.ndarray,
        sampling_rate: int,
    ) -> np.ndarray:
        max_peak_amplitude = np.max(np.abs(samples))

        # Treat signals with extremely low amplitude (below -96 dBFS peak) as silence
        if max_peak_amplitude == 0 or 20 * np.log10(max_peak_amplitude) < -96:
            return samples.copy()

        # Normalize to 0 dBFS
        if self.normalize:
            samples = samples / max_peak_amplitude

        # Apply gain
        if abs(self.gain_db) >= 0.1:
            gain_linear = 10 ** (self.gain_db / 20.0)
            samples = samples * gain_linear

        if self.hard:
            saturated_samples = np.clip(samples, -1.0, 1.0)
        else:
            saturated_samples = np.tanh(samples)

        # Revert the additional gain applied before clipping
        if abs(self.gain_db) >= 0.1:
            saturated_samples = saturated_samples / gain_linear

        # Revert 0 dBFS normalization
        if self.normalize:
            saturated_samples = saturated_samples * max_peak_amplitude

        return saturated_samples.copy()

    def reverse_timestamps(self, offset, duration, sampling_rate):
        """
        Clipping doesn't change the timing of the audio, so timestamps remain unchanged.
        """
        return offset, duration



# NOTE: The code below originates from torchaudio repository, version 2.6.0.
#       It can be found under: https://github.com/pytorch/audio/tree/release/2.6
#       The modifications applied are mostly cosmetic.
#       The inclusion of this code in lhotse allows us to perform very fast CPU resampling
#       inside the dataloader without introducing a problematic dependency on torchaudio.
#       This code is licensed under the BSD 2-Clause License,
#       included verbatim from the torchaudio repository below:
#
# BSD 2-Clause License
#
# Copyright (c) 2017 Facebook Inc. (Soumith Chintala),
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import math
from typing import Optional

import torch


class Resample(torch.nn.Module):
    r"""Resample a signal from one frequency to another. A resampling method can be given.

    .. devices:: CPU CUDA

    .. properties:: Autograd TorchScript

    Note:
        If resampling on waveforms of higher precision than float32, there may be a small loss of precision
        because the kernel is cached once as float32. If high precision resampling is important for your application,
        the functional form will retain higher precision, but run slower because it does not cache the kernel.
        Alternatively, you could rewrite a transform that caches a higher precision kernel.

    Args:
        orig_freq (int, optional): The original frequency of the signal. (Default: ``16000``)
        new_freq (int, optional): The desired frequency. (Default: ``16000``)
        resampling_method (str, optional): The resampling method to use.
            Options: [``sinc_interp_hann``, ``sinc_interp_kaiser``] (Default: ``"sinc_interp_hann"``)
        lowpass_filter_width (int, optional): Controls the sharpness of the filter, more == sharper
            but less efficient. (Default: ``6``)
        rolloff (float, optional): The roll-off frequency of the filter, as a fraction of the Nyquist.
            Lower values reduce anti-aliasing, but also reduce some of the highest frequencies. (Default: ``0.99``)
        beta (float or None, optional): The shape parameter used for kaiser window.
        dtype (torch.device, optional):
            Determnines the precision that resampling kernel is pre-computed and cached. If not provided,
            kernel is computed with ``torch.float64`` then cached as ``torch.float32``.
            If you need higher precision, provide ``torch.float64``, and the pre-computed kernel is computed and
            cached as ``torch.float64``. If you use resample with lower precision, then instead of providing this
            providing this argument, please use ``Resample.to(dtype)``, so that the kernel generation is still
            carried out on ``torch.float64``.

    Example
        >>> waveform, sample_rate = ...
        >>> transform = Resample(sample_rate, sample_rate/10)
        >>> waveform = transform(waveform)
    """

    def __init__(
        self,
        orig_freq: int = 16000,
        new_freq: int = 16000,
        resampling_method: str = "sinc_interp_hann",
        lowpass_filter_width: int = 6,
        rolloff: float = 0.99,
        beta: Optional[float] = None,
        *,
        dtype: Optional[torch.dtype] = None,
    ) -> None:
        super().__init__()

        self.orig_freq = orig_freq
        self.new_freq = new_freq
        self.gcd = math.gcd(int(self.orig_freq), int(self.new_freq))
        self.resampling_method = resampling_method
        self.lowpass_filter_width = lowpass_filter_width
        self.rolloff = rolloff
        self.beta = beta

        if self.orig_freq != self.new_freq:
            kernel, self.width = _get_sinc_resample_kernel(
                self.orig_freq,
                self.new_freq,
                self.gcd,
                self.lowpass_filter_width,
                self.rolloff,
                self.resampling_method,
                beta,
                dtype=dtype,
            )
            self.register_buffer("kernel", kernel)

    def forward(self, waveform: torch.Tensor) -> torch.Tensor:
        r"""
        Args:
            waveform (Tensor): Tensor of audio of dimension (..., time).

        Returns:
            Tensor: Output signal of dimension (..., time).
        """
        if self.orig_freq == self.new_freq:
            return waveform
        return _apply_sinc_resample_kernel(
            waveform, self.orig_freq, self.new_freq, self.gcd, self.kernel, self.width
        )


def resample(
    waveform: torch.Tensor,
    orig_freq: int,
    new_freq: int,
    lowpass_filter_width: int = 6,
    rolloff: float = 0.99,
    resampling_method: str = "sinc_interp_hann",
    beta: Optional[float] = None,
) -> torch.Tensor:
    r"""Resamples the waveform at the new frequency using bandlimited interpolation. :cite:`RESAMPLE`.

    .. devices:: CPU CUDA

    .. properties:: Autograd TorchScript

    Note:
        ``Resample`` precomputes and reuses the resampling kernel, so using it will result in
        more efficient computation if resampling multiple waveforms with the same resampling parameters.

    Args:
        waveform (Tensor): The input signal of dimension `(..., time)`
        orig_freq (int): The original frequency of the signal
        new_freq (int): The desired frequency
        lowpass_filter_width (int, optional): Controls the sharpness of the filter, more == sharper
            but less efficient. (Default: ``6``)
        rolloff (float, optional): The roll-off frequency of the filter, as a fraction of the Nyquist.
            Lower values reduce anti-aliasing, but also reduce some of the highest frequencies. (Default: ``0.99``)
        resampling_method (str, optional): The resampling method to use.
            Options: [``"sinc_interp_hann"``, ``"sinc_interp_kaiser"``] (Default: ``"sinc_interp_hann"``)
        beta (float or None, optional): The shape parameter used for kaiser window.

    Returns:
        Tensor: The waveform at the new frequency of dimension `(..., time).`
    """

    if orig_freq <= 0.0 or new_freq <= 0.0:
        raise ValueError("Original frequency and desired frequecy should be positive")

    if orig_freq == new_freq:
        return waveform

    gcd = math.gcd(int(orig_freq), int(new_freq))

    kernel, width = _get_sinc_resample_kernel(
        orig_freq,
        new_freq,
        gcd,
        lowpass_filter_width,
        rolloff,
        resampling_method,
        beta,
        waveform.device,
        waveform.dtype,
    )
    resampled = _apply_sinc_resample_kernel(
        waveform, orig_freq, new_freq, gcd, kernel, width
    )
    return resampled


def _get_sinc_resample_kernel(
    orig_freq: int,
    new_freq: int,
    gcd: int,
    lowpass_filter_width: int = 6,
    rolloff: float = 0.99,
    resampling_method: str = "sinc_interp_hann",
    beta: Optional[float] = None,
    device: torch.device = "cpu",
    dtype: Optional[torch.dtype] = None,
):
    if not (int(orig_freq) == orig_freq and int(new_freq) == new_freq):
        raise Exception(
            "Frequencies must be of integer type to ensure quality resampling computation. "
            "To work around this, manually convert both frequencies to integer values "
            "that maintain their resampling rate ratio before passing them into the function. "
            "Example: To downsample a 44100 hz waveform by a factor of 8, use "
            "`orig_freq=8` and `new_freq=1` instead of `orig_freq=44100` and `new_freq=5512.5`. "
            "For more information, please refer to https://github.com/pytorch/audio/issues/1487."
        )

    if resampling_method not in ["sinc_interp_hann", "sinc_interp_kaiser"]:
        raise ValueError("Invalid resampling method: {}".format(resampling_method))

    orig_freq = int(orig_freq) // gcd
    new_freq = int(new_freq) // gcd

    if lowpass_filter_width <= 0:
        raise ValueError("Low pass filter width should be positive.")
    base_freq = min(orig_freq, new_freq)
    # This will perform antialiasing filtering by removing the highest frequencies.
    # At first I thought I only needed this when downsampling, but when upsampling
    # you will get edge artifacts without this, as the edge is equivalent to zero padding,
    # which will add high freq artifacts.
    base_freq *= rolloff

    # The key idea of the algorithm is that x(t) can be exactly reconstructed from x[i] (tensor)
    # using the sinc interpolation formula:
    #   x(t) = sum_i x[i] sinc(pi * orig_freq * (i / orig_freq - t))
    # We can then sample the function x(t) with a different sample rate:
    #    y[j] = x(j / new_freq)
    # or,
    #    y[j] = sum_i x[i] sinc(pi * orig_freq * (i / orig_freq - j / new_freq))

    # We see here that y[j] is the convolution of x[i] with a specific filter, for which
    # we take an FIR approximation, stopping when we see at least `lowpass_filter_width` zeros crossing.
    # But y[j+1] is going to have a different set of weights and so on, until y[j + new_freq].
    # Indeed:
    # y[j + new_freq] = sum_i x[i] sinc(pi * orig_freq * ((i / orig_freq - (j + new_freq) / new_freq))
    #                 = sum_i x[i] sinc(pi * orig_freq * ((i - orig_freq) / orig_freq - j / new_freq))
    #                 = sum_i x[i + orig_freq] sinc(pi * orig_freq * (i / orig_freq - j / new_freq))
    # so y[j+new_freq] uses the same filter as y[j], but on a shifted version of x by `orig_freq`.
    # This will explain the F.conv1d after, with a stride of orig_freq.
    width = math.ceil(lowpass_filter_width * orig_freq / base_freq)
    # If orig_freq is still big after GCD reduction, most filters will be very unbalanced, i.e.,
    # they will have a lot of almost zero values to the left or to the right...
    # There is probably a way to evaluate those filters more efficiently, but this is kept for
    # future work.
    idx_dtype = dtype if dtype is not None else torch.float64

    idx = (
        torch.arange(-width, width + orig_freq, dtype=idx_dtype, device=device)[
            None, None
        ]
        / orig_freq
    )

    t = (
        torch.arange(0, -new_freq, -1, dtype=dtype, device=device)[:, None, None]
        / new_freq
        + idx
    )
    t *= base_freq
    t = t.clamp_(-lowpass_filter_width, lowpass_filter_width)

    # we do not use built in torch windows here as we need to evaluate the window
    # at specific positions, not over a regular grid.
    if resampling_method == "sinc_interp_hann":
        window = torch.cos(t * math.pi / lowpass_filter_width / 2) ** 2
    else:
        # sinc_interp_kaiser
        if beta is None:
            beta = 14.769656459379492
        beta_tensor = torch.tensor(float(beta))
        window = torch.i0(
            beta_tensor * torch.sqrt(1 - (t / lowpass_filter_width) ** 2)
        ) / torch.i0(beta_tensor)

    t *= math.pi

    scale = base_freq / orig_freq
    kernels = torch.where(t == 0, torch.tensor(1.0).to(t), t.sin() / t)
    kernels *= window * scale

    if dtype is None:
        kernels = kernels.to(dtype=torch.float32)

    return kernels, width


def _apply_sinc_resample_kernel(
    waveform: torch.Tensor,
    orig_freq: int,
    new_freq: int,
    gcd: int,
    kernel: torch.Tensor,
    width: int,
):
    if not waveform.is_floating_point():
        raise TypeError(
            f"Expected floating point type for waveform tensor, but received {waveform.dtype}."
        )

    orig_freq = int(orig_freq) // gcd
    new_freq = int(new_freq) // gcd

    # pack batch
    shape = waveform.size()
    waveform = waveform.view(-1, shape[-1])

    num_wavs, length = waveform.shape
    waveform = torch.nn.functional.pad(waveform, (width, width + orig_freq))
    resampled = torch.nn.functional.conv1d(waveform[:, None], kernel, stride=orig_freq)
    resampled = resampled.transpose(1, 2).reshape(num_wavs, -1)
    target_length = torch.ceil(torch.as_tensor(new_freq * length / orig_freq)).long()
    resampled = resampled[..., :target_length]

    # unpack batch
    resampled = resampled.view(shape[:-1] + resampled.shape[-1:])
    return resampled



from dataclasses import dataclass, field
from typing import Callable, List, Optional, Tuple, Union

import numpy as np
import torch

from lhotse.augmentation.transform import AudioTransform
from lhotse.augmentation.utils import FastRandomRIRGenerator, convolve1d
from lhotse.utils import Seconds


@dataclass
class ReverbWithImpulseResponse(AudioTransform):
    """
    Reverberation effect by convolving with a room impulse response.
    This code is based on Kaldi's wav-reverberate utility:
    https://github.com/kaldi-asr/kaldi/blob/master/src/featbin/wav-reverberate.cc
    If no ``rir_recording`` is provided, we will generate an impulse response using a fast random
    generator (https://arxiv.org/abs/2208.04101).
    The impulse response can possibly be multi-channel, in which case multi-channel reverberated
    audio can be obtained by appropriately setting `rir_channels`. For example, `rir_channels=[0,1]`
    will convolve using the first two channels of the impulse response, generating a stereo
    reverberated audio.
    Note that we enforce the --shift-output option in Kaldi's wav-reverberate utility,
    which means that the output length will be equal to the input length.
    """

    rir: Optional[dict] = None
    normalize_output: bool = True
    early_only: bool = False
    rir_channels: List[int] = field(default_factory=lambda: [0])
    rir_generator: Optional[Union[dict, Callable]] = None

    RIR_SCALING_FACTOR: float = 0.5**15

    def __post_init__(self):
        if isinstance(self.rir, dict):
            from lhotse.serialization import deserialize_item

            # Pass a shallow copy of the RIR dict since deserialization is destructive
            # If RIR is a Cut, we have to perform one extra copy (hacky but better than deepcopy).
            rir = self.rir.copy()
            if "recording" in self.rir:
                rir["recording"] = rir["recording"].copy()
            self.rir = deserialize_item(rir)

        assert (
            self.rir is not None or self.rir_generator is not None
        ), "Either `rir` or `rir_generator` must be provided."

        if self.rir is not None:
            assert all(
                c < self.rir.num_channels for c in self.rir_channels
            ), "Invalid channel index in `rir_channels`"

        if self.rir_generator is not None and isinstance(self.rir_generator, dict):
            self.rir_generator = FastRandomRIRGenerator(**self.rir_generator)

    def to_dict(self) -> dict:
        from lhotse import Recording
        from lhotse.cut import Cut

        return {
            "name": type(self).__name__,
            "kwargs": {
                "rir": self.rir.to_dict()
                if isinstance(self.rir, (Recording, Cut))
                else self.rir,
                "normalize_output": self.normalize_output,
                "early_only": self.early_only,
                "rir_channels": list(self.rir_channels),
                "rir_generator": self.rir_generator
                if self.rir_generator is None or isinstance(self.rir_generator, dict)
                else self.rir_generator.to_dict(),
            },
        }

    def __call__(
        self,
        samples: np.ndarray,
        sampling_rate: int,
    ) -> np.ndarray:
        """
        :param samples: The audio samples to reverberate.
        :param sampling_rate: The sampling rate of the audio samples.
        """
        D_in, N_in = samples.shape
        input_is_mono = D_in == 1

        # The following cases are possible:
        # Case 1: input is mono, rir is mono -> mono output
        #   We will generate a random mono rir if not provided explicitly.
        # Case 2: input is mono, rir is multi-channel -> multi-channel output
        #   This requires a user-provided rir, since we cannot simulate array microphone.
        # Case 3: input is multi-channel, rir is mono -> multi-channel output
        #   This does not make much sense, but we will apply the same rir to all channels.
        # 4. input is multi-channel, rir is multi-channel -> multi-channel output
        #   This also requires a user-provided rir. Also, the number of channels in the rir
        #   must match the number of channels in the input.

        # Let us make some assertions based on the above.
        if input_is_mono:
            assert (
                self.rir is not None or len(self.rir_channels) == 1
            ), "For mono input, either provide an RIR explicitly or set rir_channels to [0]."
        else:
            assert len(self.rir_channels) == 1 or len(self.rir_channels) == D_in, (
                "For multi-channel input, we only support mono RIR or RIR with the same number "
                "of channels as the input."
            )

        # Generate a random RIR if not provided.
        if self.rir is None:
            rir_ = self.rir_generator(nsource=1)
        else:
            from lhotse import Recording

            rir = self.rir.to_cut() if isinstance(self.rir, Recording) else self.rir
            rir = rir.with_channels(self.rir_channels)
            if self.early_only:
                rir = rir.truncate(duration=0.05)
            rir_ = rir.load_audio()

        D_rir, N_rir = rir_.shape
        N_out = N_in  # Enforce shift output
        # output is multi-channel if either input or rir is multi-channel
        D_out = D_rir if input_is_mono else D_in

        # if RIR is mono, repeat it to match the number of channels in the input
        rir_ = rir_.repeat(D_out, axis=0) if D_rir == 1 else rir_

        # Initialize output matrix with the specified input channel.
        augmented = np.zeros((D_out, N_out), dtype=samples.dtype)

        for d in range(D_out):
            d_in = 0 if input_is_mono else d
            augmented[d, :N_in] = samples[d_in]
            power_before_reverb = np.sum(np.abs(samples[d_in]) ** 2) / N_in
            rir_d = rir_[d, :] * self.RIR_SCALING_FACTOR

            # Convolve the signal with impulse response.
            aug_d = convolve1d(
                torch.from_numpy(samples[d_in]), torch.from_numpy(rir_d)
            ).numpy()
            shift_index = np.argmax(rir_d)
            augmented[d, :] = aug_d[shift_index : shift_index + N_out]

            if self.normalize_output:
                power_after_reverb = np.sum(np.abs(augmented[d, :]) ** 2) / N_out
                if power_after_reverb > 0:
                    augmented[d, :] *= np.sqrt(power_before_reverb / power_after_reverb)

        return augmented

    def reverse_timestamps(
        self,
        offset: Seconds,
        duration: Optional[Seconds],
        sampling_rate: Optional[int],  # Not used, made for compatibility purposes
    ) -> Tuple[Seconds, Optional[Seconds]]:
        """
        This method just returns the original offset and duration since we have
        implemented output shifting which preserves these properties.
        """

        return offset, duration



import warnings
from functools import partial
from typing import Dict, List, Literal, Optional, Tuple, Type, Union

from lhotse import fastcopy
from lhotse.array import Array, TemporalArray
from lhotse.audio import Recording
from lhotse.cut import Cut
from lhotse.shar.utils import to_shar_placeholder
from lhotse.shar.writers.array import ArrayTarWriter
from lhotse.shar.writers.audio import AudioTarWriter
from lhotse.shar.writers.cut import JsonlShardWriter
from lhotse.utils import Pathlike, ifnone

WriterName = Literal["wav", "flac", "mp3", "lilcom", "numpy", "jsonl"]
FieldWriterInstance = Union[AudioTarWriter, ArrayTarWriter]
FieldWriter = Type[FieldWriterInstance]


class SharWriter:
    """
    SharWriter writes cuts and their corresponding data into multiple shards,
    also recognized as the Lhotse Shar format.
    Each shard is numbered and represented as a collection of one text manifest and
    one or more binary tarfiles.
    Each tarfile contains a single type of data, e.g., recordings, features, or custom fields.

    The main idea behind Lhotse Shar format is to optimize dataloading with sequential reads,
    while keeping the data composition more flexible than e.g. WebDataset tar archives do.
    To achieve this, Lhotse Shar keeps each data type in a separate archive, along a single
    CutSet JSONL manifest.
    This way, the metadata can be investigated without iterating through the binary data.
    The format also allows iteration over a subset of fields, or extension of existing data
    with new fields.

    The user has to specify which fields should be saved, and what compression to use for each of them.
    Currently we support ``wav``, ``flac``, ``opus``, and ``mp3`` compression for ``recording`` and custom audio fields,
    and ``lilcom`` or ``numpy`` for ``features`` and custom array fields.

    Example::

        >>> cuts = CutSet(...)  # cuts have 'recording' and 'features'
        >>> with SharWriter("some_dir", shard_size=100, fields={"recording": "opus", "features": "lilcom"}) as w:
        ...     for cut in cuts:
        ...         w.write(cut)

    .. note:: Different audio backends in Lhotse may use different encoders for the same audio formats.
        It is advisable to use the same audio backend for saving and loading audio data in Shar and other formats.
        See: :class:`lhotse.audio.recording.Recording`.

    It would create a directory ``some_dir`` with files such as ``some_dir/cuts.000000.jsonl.gz``,
    ``some_dir/recording.000000.tar``, ``some_dir/features.000000.tar``,
    and then the same names but numbered with ``000001``, etc.
    The starting shard offset can be set using ``shard_offset`` parameter. The writer starts from 0 by default.

    When ``shard_size`` is set to ``None``, we will disable automatic sharding and the
    shard number suffix will be omitted from the file names.

    The option ``warn_unused_fields`` will emit a warning when cuts have some data attached to them
    (e.g., recording, features, or custom arrays) but saving it was not specified via ``fields``.

    The option ``include_cuts`` controls whether we store the cuts alongside ``fields`` (true by default).
    Turning it off is useful when extending existing dataset with new fields/feature types,
    but the original cuts do not require any modification.

    See also: :class:`~lhotse.shar.writers.tar.TarWriter`, :class:`~lhotse.shar.writers.audio.AudioTarWriter`,
        :class:`~lhotse.shar.writers.array.ArrayTarWriter`.
    """

    def __init__(
        self,
        output_dir: Pathlike,
        fields: Dict[str, str],
        shard_size: Optional[int] = 1000,
        warn_unused_fields: bool = True,
        include_cuts: bool = True,
        shard_suffix: Optional[str] = None,
        shard_offset: int = 0,
    ) -> None:
        self.output_dir = str(output_dir)
        self.shard_size = shard_size
        self.fields = fields
        self.warn_unused_fields = warn_unused_fields
        self.include_cuts = include_cuts
        if self.sharding_enabled:
            assert (
                shard_suffix is None
            ), f"shard_suffix must be None when shard_size is specified (got: '{shard_suffix}')."
            self.shard_suffix = ".%06d"
        else:
            self.shard_suffix = ifnone(shard_suffix, "")
        self.initial_shard_offset = shard_offset

        self.writers = {}
        if include_cuts:
            self.writers["cuts"] = JsonlShardWriter(
                pattern=_create_cuts_output_url(self.output_dir, self.shard_suffix),
                shard_size=self.shard_size,
                shard_offset=self.initial_shard_offset,
            )
        for field, writer_type in self.fields.items():
            make_writer_fn, ext = resolve_writer(writer_type)
            self.writers[field] = make_writer_fn(
                pattern=f"{self.output_dir}/{field}{self.shard_suffix}{ext}",
                shard_size=self.shard_size,
                shard_offset=self.initial_shard_offset,
            )

    @property
    def sharding_enabled(self) -> bool:
        return self.shard_size is not None and self.shard_size > 0

    @property
    def output_paths(self) -> Dict[str, List[str]]:
        return {k: w.output_paths for k, w in self.writers.items()}

    def __enter__(self):
        for w in self.writers.values():
            w.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        for w in self.writers.values():
            w.close()

    def write(self, cut: Cut) -> None:

        # Handle audio.
        if "recording" in self.fields:
            if cut.has_recording:
                data = cut.load_audio()
                recording = to_shar_placeholder(cut.recording, cut)
                cut_channels = _aslist(cut.channel)
                if recording.channel_ids != cut_channels:
                    # If recording is multi-channel but the cut refers to a subset of them,
                    # we have to update the recording manifest accordingly
                    recording.sources[0].channels = cut_channels
                    recording.channel_ids = cut_channels
                self.writers["recording"].write(
                    cut.id,
                    data,
                    cut.sampling_rate,
                    manifest=recording,
                    original_format=cut.recording.source_format,
                )
                cut = fastcopy(cut, recording=recording)
            else:
                self.writers["recording"].write_placeholder(cut.id)
        elif cut.has_recording and self.warn_unused_fields:
            warnings.warn(
                "Found cut with 'recording' field that is not specified for Shar writing."
            )

        # Handle features.
        if "features" in self.fields:
            if cut.has_features:
                data = cut.load_features()
                features = to_shar_placeholder(cut.features, cut)
                self.writers["features"].write(cut.id, data, manifest=features)
                cut = fastcopy(cut, features=features)
            else:
                self.writers["features"].write_placeholder(cut.id)
        elif cut.has_features and self.warn_unused_fields:
            warnings.warn(
                "Found cut with 'features' field that is not specified for Shar writing."
            )

        # Handle custom data and non-data attributes.
        for key in self.fields:

            # Skip fields already taken care of.
            if key in ["recording", "features"]:
                continue

            # Check if the custom attribute is available: if yes
            if cut.has_custom(key):
                val = getattr(cut, key)
                if not isinstance(val, (Array, TemporalArray, Recording)):
                    assert isinstance(
                        self.writers[key], JsonlShardWriter
                    ), f"Expected writer type 'jsonl' (got '{self.fields[key]}') for non-data field '{key}'."
                    self.writers[key].write({"cut_id": cut.id, key: val})
                else:
                    data = cut.load_custom(key)
                    placeholder_obj = to_shar_placeholder(val, cut)
                    channel_selector_key = f"{key}_channel_selector"
                    kwargs = {}
                    if isinstance(val, Recording):
                        kwargs["sampling_rate"] = val.sampling_rate
                        if cut.has_custom(channel_selector_key):
                            # override custom recording channels since the audio was loaded via cut
                            # and used the channel selector
                            placeholder_obj.sources[0].channels = cut.custom[
                                channel_selector_key
                            ]
                            placeholder_obj.channel_ids = cut.custom[
                                channel_selector_key
                            ]
                    self.writers[key].write(
                        cut.id, data, manifest=placeholder_obj, **kwargs
                    )
                    cut = fastcopy(cut, custom=cut.custom.copy())
                    cut.custom.pop(channel_selector_key, None)  # no longer needed
                    setattr(cut, key, placeholder_obj)
            else:
                self.writers[key].write_placeholder(cut.id)

        # Warn in case there is some data that wasn't requested to be saved.
        for key, val in ifnone(cut.custom, {}).items():
            if not isinstance(val, (Array, TemporalArray, Recording)):
                continue
            if key not in self.fields:
                if self.warn_unused_fields:
                    warnings.warn(
                        f"Found cut with '{key}' field that is not specified for Shar writing."
                    )
                continue

        # We will write only the relevant subset of the binary data alongside cut,
        # so we need to update the offset (start).
        cut = fastcopy(cut, start=0)

        if "cuts" in self.writers:
            self.writers["cuts"].write(cut)


def resolve_writer(name: str) -> Tuple[FieldWriter, str]:
    opts = {
        "wav": (partial(AudioTarWriter, format="wav"), ".tar"),
        "flac": (partial(AudioTarWriter, format="flac"), ".tar"),
        "mp3": (partial(AudioTarWriter, format="mp3"), ".tar"),
        "opus": (partial(AudioTarWriter, format="opus"), ".tar"),
        "original": (partial(AudioTarWriter, format="original"), ".tar"),
        "lilcom": (partial(ArrayTarWriter, compression="lilcom"), ".tar"),
        "numpy": (partial(ArrayTarWriter, compression="numpy"), ".tar"),
        "jsonl": (JsonlShardWriter, ".jsonl.gz"),
    }
    assert (
        name in opts
    ), f"Unknown field type (got: '{name}', we support only: {', '.join(opts)}"
    return opts[name]


def _create_cuts_output_url(base_output_url: str, shard_suffix: str) -> str:

    # special case where we want to ensure the CutSet actually gets gzipped
    if base_output_url.startswith("pipe:"):
        base_output_url = base_output_url.replace("pipe:", "pipe:gzip -c | ")

    return f"{base_output_url}/cuts{shard_suffix}.jsonl.gz"


def _aslist(x):
    if isinstance(x, list):
        return x
    return [x]
